# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import unittest

from azure.cli.core.aaz import AAZIdentityObjectType
from azure.cli.core.aaz._field_value import AAZIdentityObject


class TestAAZIdentity(unittest.TestCase):
    identity = AAZIdentityObject(schema=AAZIdentityObjectType(), data={})

    def test_main_command(self):
        result = {}
        data = {
            "userAssigned": ["a", "b"],
            "systemAssigned": "SystemAssigned",
            "action": "create",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "SystemAssigned,UserAssigned")
        self.assertTrue(result["userAssignedIdentities"] == {"a": {}, "b": {}})

        result = {}
        data = {
            "userAssigned": [],
            "systemAssigned": "SystemAssigned",
            "action": "create",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "SystemAssigned")
        self.assertTrue("userAssignedIdentities" not in result)

        result = {}
        data = {
            "userAssigned": [],
            "action": "create",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result == {})

        result = {}
        data = {"action": "create"}
        result = self.identity._build_identity(data, result)
        self.assertTrue(result == {})

    def test_assign_command(self):
        result = {
            "type": "systemAssigned, userAssigned",
            "userAssignedIdentities": {"a": {}, "b": {}},
        }
        data = {
            "userAssigned": ["b", "c"],
            "action": "assign",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "SystemAssigned,UserAssigned")
        self.assertTrue(result["userAssignedIdentities"] == {"a": {}, "b": {}, "c": {}})

        result = {
            "type": "userAssigned",
            "userAssignedIdentities": {"a": {}, "b": {}},
        }
        data = {
            "systemAssigned": "SystemAssigned",
            "action": "assign",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "SystemAssigned,UserAssigned")
        self.assertTrue(result["userAssignedIdentities"] == {"a": {}, "b": {}})

        result = {
            "type": "systemAssigned",
        }
        data = {
            "userAssigned": ["a", "b"],
            "action": "assign",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "SystemAssigned,UserAssigned")
        self.assertTrue(result["userAssignedIdentities"] == {"a": {}, "b": {}})

        result = {}
        data = {
            "userAssigned": ["a", "b"],
            "action": "assign",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "UserAssigned")
        self.assertTrue(result["userAssignedIdentities"] == {"a": {}, "b": {}})

        result = {}
        data = {
            "userAssigned": [],
            "action": "assign",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result == {})

    def test_remove_command(self):
        result = {
            "type": "systemAssigned, userAssigned",
            "userAssignedIdentities": {"a": {}, "b": {}},
        }
        data = {
            "userAssigned": ["b", "c"],
            "systemAssigned": "SystemAssigned",
            "action": "remove",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "UserAssigned")
        self.assertTrue(result["userAssignedIdentities"] == {"a": {}})

        result = {
            "type": "systemAssigned, userAssigned",
            "userAssignedIdentities": {"a": {}, "b": {}},
        }
        data = {
            "userAssigned": [],
            "action": "remove",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result["type"] == "SystemAssigned")
        self.assertTrue("userAssignedIdentities" not in result)

        result = {
            "type": "userAssigned",
            "userAssignedIdentities": {"a": {}, "b": {}},
        }
        data = {
            "userAssigned": [],
            "action": "remove",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result == {})

        result = {}
        data = {
            "userAssigned": ["a", "b"],
            "action": "remove",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result == {})

        result = {}
        data = {
            "userAssigned": [],
            "action": "remove",
        }
        result = self.identity._build_identity(data, result)
        self.assertTrue(result == {})



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import re
from ipaddress import ip_network
from math import isclose, isnan

from azure.mgmt.containerservice.models import KubernetesSupportPlan
from azure.cli.command_modules.acs._consts import (
    CONST_MANAGED_CLUSTER_SKU_TIER_FREE,
    CONST_MANAGED_CLUSTER_SKU_TIER_STANDARD,
    CONST_MANAGED_CLUSTER_SKU_TIER_PREMIUM,
    CONST_OS_SKU_AZURELINUX,
    CONST_OS_SKU_CBLMARINER,
    CONST_OS_SKU_MARINER,
    CONST_NETWORK_POD_IP_ALLOCATION_MODE_DYNAMIC_INDIVIDUAL,
    CONST_NETWORK_POD_IP_ALLOCATION_MODE_STATIC_BLOCK,
    CONST_NODEPOOL_MODE_GATEWAY,
    CONST_AZURE_SERVICE_MESH_MAX_EGRESS_NAME_LENGTH,
    CONST_VIRTUAL_MACHINE_SCALE_SETS,
    CONST_AVAILABILITY_SET,
    CONST_VIRTUAL_MACHINES,
)
from azure.cli.core import keys
from azure.cli.core.azclierror import (
    ArgumentUsageError,
    InvalidArgumentValueError,
    MutuallyExclusiveArgumentError,
    RequiredArgumentMissingError,
)

from azure.cli.core.commands.validators import validate_tag
from azure.cli.core.util import CLIError
from knack.log import get_logger


logger = get_logger(__name__)


def validate_ssh_key(namespace):
    if hasattr(namespace, 'no_ssh_key') and namespace.no_ssh_key:
        return
    string_or_file = (namespace.ssh_key_value or
                      os.path.join(os.path.expanduser('~'), '.ssh', 'id_rsa.pub'))
    content = string_or_file
    if os.path.exists(string_or_file):
        logger.info('Use existing SSH public key file: %s', string_or_file)
        with open(string_or_file, 'r') as f:
            content = f.read()
    elif not keys.is_valid_ssh_rsa_public_key(content):
        if namespace.generate_ssh_keys:
            # figure out appropriate file names:
            # 'base_name'(with private keys), and 'base_name.pub'(with public keys)
            public_key_filepath = string_or_file
            if public_key_filepath[-4:].lower() == '.pub':
                private_key_filepath = public_key_filepath[:-4]
            else:
                private_key_filepath = public_key_filepath + '.private'
            content = keys.generate_ssh_keys(private_key_filepath, public_key_filepath)
            logger.warning("SSH key files '%s' and '%s' have been generated under ~/.ssh to "
                           "allow SSH access to the VM. If using machines without "
                           "permanent storage like Azure Cloud Shell without an attached "
                           "file share, back up your keys to a safe location",
                           private_key_filepath, public_key_filepath)
        else:
            if (not content or str(content).strip() == "" or
                    (content == os.path.join(os.path.expanduser('~'), '.ssh', 'id_rsa.pub'))):
                namespace.no_ssh_key = True
                return
            raise CLIError(
                "The SSH key provided is not a valid RSA public key. "
                "Provide the contents of a valid SSH public key (for example, '~/.ssh/id_rsa.pub'), "
                "specify a path to a public key file, "
                "or add --generate-ssh-keys as a parameter to create a new key pair."
            )
    namespace.ssh_key_value = content


def validate_create_parameters(namespace):
    if not namespace.name:
        raise CLIError('--name has no value')
    if namespace.dns_name_prefix is not None and not namespace.dns_name_prefix:
        raise CLIError('--dns-prefix has no value')


def validate_ip_ranges(namespace):
    if not namespace.api_server_authorized_ip_ranges:
        return

    restrict_traffic_to_agentnodes = "0.0.0.0/32"
    allow_all_traffic = ""
    ip_ranges = [ip.strip() for ip in namespace.api_server_authorized_ip_ranges.split(",")]

    if restrict_traffic_to_agentnodes in ip_ranges and len(ip_ranges) > 1:
        raise CLIError("Setting --api-server-authorized-ip-ranges to 0.0.0.0/32 is not allowed with other IP ranges."
                       "Refer to https://aka.ms/aks/whitelist for more details")

    if allow_all_traffic in ip_ranges and len(ip_ranges) > 1:
        raise CLIError("--api-server-authorized-ip-ranges cannot be disabled and simultaneously enabled")

    for ip in ip_ranges:
        if ip in [restrict_traffic_to_agentnodes, allow_all_traffic]:
            continue
        try:
            ip = ip_network(ip)
            if not ip.is_global:
                raise CLIError("--api-server-authorized-ip-ranges must be global non-reserved addresses or CIDRs")
            if ip.version == 6:
                raise CLIError("--api-server-authorized-ip-ranges cannot be IPv6 addresses")
        except ValueError:
            raise CLIError("--api-server-authorized-ip-ranges should be a list of IPv4 addresses or CIDRs")


def validate_namespace_name(namespace):
    _validate_namespace_name(namespace.name)


def _validate_namespace_name(name):
    """
    Validates a Kubernetes namespace name.
    Raises ValueError if the name is invalid.
    """
    if name != "":
        if len(name) < 1 or len(name) > 63:
            raise ValueError("Namespace name must be between 1 and 63 characters.")
        pattern = r'^[a-z0-9]([-a-z0-9]*[a-z0-9])?$'
        if not re.match(pattern, name):
            raise ValueError(
                f"Invalid namespace '{name}'. Must consist of lower case alphanumeric characters or '-', "
                "and must start and end with an alphanumeric character."
            )


def validate_resource_quota(namespace):
    if namespace.cpu_request is not None:
        if not namespace.cpu_request.endswith("m"):
            raise ValueError("--cpu-request must be specified in millicores, like 200m")
    if namespace.cpu_limit is not None:
        if not namespace.cpu_limit.endswith("m"):
            raise ValueError("--cpu-limit must be specified in millicores, like 200m")
    pattern = r"^\d+(Ki|Mi|Gi|Ti|Pi|Ei)$"
    if namespace.memory_request is not None:
        if not re.match(pattern, namespace.memory_request):
            raise ValueError("--memory-request must be specified in the power-of-two equivalents form:"
                             "Ei, Pi, Ti, Gi, Mi, Ki.")
    if namespace.memory_limit is not None:
        if not re.match(pattern, namespace.memory_limit):
            raise ValueError("--memory-limit must be specified in the power-of-two equivalents form:"
                             "Ei, Pi, Ti, Gi, Mi, Ki.")


def validate_k8s_version(namespace):
    """Validates a string as a possible Kubernetes version. An empty string is also valid, which tells the server
    to use its default version."""
    if namespace.kubernetes_version:
        k8s_release_regex = re.compile(r'^[v|V]?(\d+\.\d+(?:\.\d+)?)$')
        found = k8s_release_regex.findall(namespace.kubernetes_version)
        if found:
            namespace.kubernetes_version = found[0]
        else:
            raise CLIError('--kubernetes-version should be the full version number or major.minor version number, '
                           'such as "1.7.12" or "1.7"')


def _validate_nodepool_name(nodepool_name):
    """Validates a nodepool name to be at most 12 characters, alphanumeric only."""
    if nodepool_name != "":
        if len(nodepool_name) > 12:
            raise InvalidArgumentValueError('--nodepool-name can contain at most 12 characters')
        if not nodepool_name.isalnum():
            raise InvalidArgumentValueError('--nodepool-name should contain only alphanumeric characters')


def validate_nodepool_name(namespace):
    """Validates a nodepool name to be at most 12 characters, alphanumeric only."""
    _validate_nodepool_name(namespace.nodepool_name)


def validate_agent_pool_name(namespace):
    """Validates a nodepool name to be at most 12 characters, alphanumeric only."""
    _validate_nodepool_name(namespace.agent_pool_name)


def validate_asm_egress_name(namespace):
    if namespace.istio_egressgateway_name is None:
        return
    name = namespace.istio_egressgateway_name
    asm_egress_name_regex = re.compile(r'^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$')
    match = asm_egress_name_regex.match(name)
    if not match or len(name) > CONST_AZURE_SERVICE_MESH_MAX_EGRESS_NAME_LENGTH:
        raise InvalidArgumentValueError(
            f"Istio egress name {name} is invalid. Name must be between 1 and "
            f"{CONST_AZURE_SERVICE_MESH_MAX_EGRESS_NAME_LENGTH} characters, must consist of lower case alphanumeric "
            "characters, '-' or '.', and must start and end with an alphanumeric character."
        )


def validate_kubectl_version(namespace):
    """Validates a string as a possible Kubernetes version."""
    k8s_release_regex = re.compile(r'^[v|V]?(\d+\.\d+\.\d+.*|latest)$')
    found = k8s_release_regex.findall(namespace.client_version)
    if found:
        namespace.client_version = found[0]
    else:
        raise CLIError('--client-version should be the full version number '
                       '(such as "1.11.8" or "1.12.6") or "latest"')


def validate_kubelogin_version(namespace):
    """Validates a string as a possible kubelogin version."""
    kubelogin_regex = re.compile(r'^[v|V]?(\d+\.\d+\.\d+.*|latest)$')
    found = kubelogin_regex.findall(namespace.kubelogin_version)
    if found:
        namespace.kubelogin_version = found[0]
    else:
        raise CLIError('--kubelogin-version should be the full version number '
                       '(such as "0.0.4") or "latest"')


def validate_linux_host_name(namespace):
    """Validates a string as a legal host name component.

    This validation will also occur server-side in the ARM API, but that may take
    a minute or two before the user sees it. So it's more user-friendly to validate
    in the CLI pre-flight.
    """
    # https://stackoverflow.com/questions/106179/regular-expression-to-match-dns-hostname-or-ip-address
    rfc1123_regex = re.compile(r'^[a-zA-Z0-9]$|^[a-zA-Z0-9][-_a-zA-Z0-9]{0,61}[a-zA-Z0-9]$')  # pylint:disable=line-too-long
    found = rfc1123_regex.findall(namespace.name)
    if not found:
        raise InvalidArgumentValueError('--name cannot exceed 63 characters and can only contain '
                                        'letters, numbers, underscores (_) or dashes (-).')


def validate_snapshot_name(namespace):
    """Validates a nodepool snapshot name to be alphanumeric and dashes."""
    rfc1123_regex = re.compile(r'^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])(\.([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9]))*$')  # pylint:disable=line-too-long
    found = rfc1123_regex.findall(namespace.snapshot_name)
    if not found:
        raise InvalidArgumentValueError('--name cannot exceed 63 characters and can only contain '
                                        'letters, numbers, or dashes (-).')


def validate_vm_set_type(namespace):
    """Validates the vm set type string."""
    if namespace.vm_set_type is not None:
        if namespace.vm_set_type == '':
            return
        if (
            namespace.vm_set_type.lower() != CONST_VIRTUAL_MACHINE_SCALE_SETS.lower() and
            namespace.vm_set_type.lower() != CONST_AVAILABILITY_SET.lower() and
            namespace.vm_set_type.lower() != CONST_VIRTUAL_MACHINES.lower()
        ):
            raise CLIError("--vm-set-type can only be VirtualMachineScaleSets or AvailabilitySet or VirtualMachines")


def validate_load_balancer_sku(namespace):
    """Validates the load balancer sku string."""
    if namespace.load_balancer_sku is not None:
        if namespace.load_balancer_sku == '':
            return
        if namespace.load_balancer_sku.lower() != "basic" and namespace.load_balancer_sku.lower() != "standard":
            raise CLIError("--load-balancer-sku can only be standard or basic")


def validate_sku_tier(namespace):
    """Validates the sku tier string."""
    if namespace.tier is not None:
        if namespace.tier == '':
            return
        if namespace.tier.lower() not in (
                CONST_MANAGED_CLUSTER_SKU_TIER_FREE, CONST_MANAGED_CLUSTER_SKU_TIER_STANDARD,
                CONST_MANAGED_CLUSTER_SKU_TIER_PREMIUM):
            raise InvalidArgumentValueError("--tier can only be free, standard and premium")


def validate_k8s_support_plan(namespace):
    if namespace.k8s_support_plan is not None:
        if namespace.k8s_support_plan == '':
            return
        if namespace.k8s_support_plan.lower() not in (
                KubernetesSupportPlan.KUBERNETES_OFFICIAL.lower(), KubernetesSupportPlan.AKS_LONG_TERM_SUPPORT.lower()):
            raise InvalidArgumentValueError("--k8s-support-plan can only be KubernetesOfficial or AKSLongTermSupport")


def validate_load_balancer_outbound_ips(namespace):
    """validate load balancer profile outbound IP ids"""
    if namespace.load_balancer_outbound_ips is not None:
        ip_id_list = [x.strip() for x in namespace.load_balancer_outbound_ips.split(',')]
        if not all(ip_id_list):
            raise CLIError("--load-balancer-outbound-ips cannot contain whitespace")


def validate_load_balancer_outbound_ip_prefixes(namespace):
    """validate load balancer profile outbound IP prefix ids"""
    if namespace.load_balancer_outbound_ip_prefixes is not None:
        ip_prefix_id_list = [x.strip() for x in namespace.load_balancer_outbound_ip_prefixes.split(',')]
        if not all(ip_prefix_id_list):
            raise CLIError("--load-balancer-outbound-ip-prefixes cannot contain whitespace")


def validate_load_balancer_outbound_ports(namespace):
    """validate load balancer profile outbound allocated ports"""
    if namespace.load_balancer_outbound_ports is not None:
        if namespace.load_balancer_outbound_ports % 8 != 0:
            raise CLIError("--load-balancer-allocated-ports must be a multiple of 8")
        if namespace.load_balancer_outbound_ports < 0 or namespace.load_balancer_outbound_ports > 64000:
            raise CLIError("--load-balancer-allocated-ports must be in the range [0,64000]")


def validate_load_balancer_idle_timeout(namespace):
    """validate load balancer profile idle timeout"""
    if namespace.load_balancer_idle_timeout is not None:
        if namespace.load_balancer_idle_timeout < 4 or namespace.load_balancer_idle_timeout > 100:
            raise CLIError("--load-balancer-idle-timeout must be in the range [4,100]")


def validate_network_policy(namespace):
    """validate network policy to be in lowercase"""
    if namespace.network_policy is not None and namespace.network_policy.islower() is False:
        raise InvalidArgumentValueError("--network-policy should be provided in lowercase")


def validate_nat_gateway_managed_outbound_ip_count(namespace):
    """validate NAT gateway profile managed outbound IP count"""
    if namespace.nat_gateway_managed_outbound_ip_count is not None:
        if namespace.nat_gateway_managed_outbound_ip_count < 1 or namespace.nat_gateway_managed_outbound_ip_count > 16:
            raise InvalidArgumentValueError("--nat-gateway-managed-outbound-ip-count must be in the range [1,16]")


def validate_nat_gateway_idle_timeout(namespace):
    """validate NAT gateway profile idle timeout"""
    if namespace.nat_gateway_idle_timeout is not None:
        if namespace.nat_gateway_idle_timeout < 4 or namespace.nat_gateway_idle_timeout > 120:
            raise InvalidArgumentValueError("--nat-gateway-idle-timeout must be in the range [4,120]")


def validate_nodes_count(namespace):
    """Validates that min_count and max_count is set between 0-1000"""
    if namespace.min_count is not None:
        if namespace.min_count < 0 or namespace.min_count > 1000:
            raise CLIError('--min-count must be in the range [0,1000]')
    if namespace.max_count is not None:
        if namespace.max_count < 0 or namespace.max_count > 1000:
            raise CLIError('--max-count must be in the range [0,1000]')


def validate_nodepool_taints(namespace):
    """Validates that provided node labels is a valid format"""

    if hasattr(namespace, 'nodepool_taints'):
        taintsStr = namespace.nodepool_taints
    else:
        taintsStr = namespace.node_taints

    if taintsStr is None or taintsStr == '':
        return

    for taint in taintsStr.split(','):
        validate_taint(taint)


def validate_taint(taint):
    """Validates that provided taint is a valid format"""

    regex = re.compile(r"^[a-zA-Z\d][\w\-\.\/]{0,252}=[a-zA-Z\d][\w\-\.]{0,62}:(NoSchedule|PreferNoSchedule|NoExecute)$")  # pylint: disable=line-too-long

    if taint == "":
        return
    found = regex.findall(taint)
    if not found:
        raise ArgumentUsageError('Invalid node taint: %s' % taint)


def validate_priority(namespace):
    """Validates the node pool priority string."""
    if namespace.priority is not None:
        if namespace.priority == '':
            return
        if namespace.priority != "Spot" and \
                namespace.priority != "Regular":
            raise CLIError("--priority can only be Spot or Regular")


def validate_eviction_policy(namespace):
    """Validates the node pool priority string."""
    if namespace.eviction_policy is not None:
        if namespace.eviction_policy == '':
            return
        if namespace.eviction_policy != "Delete" and \
                namespace.eviction_policy != "Deallocate":
            raise CLIError("--eviction-policy can only be Delete or Deallocate")


def validate_spot_max_price(namespace):
    """Validates the spot node pool max price."""
    if not isnan(namespace.spot_max_price):
        if namespace.priority != "Spot":
            raise CLIError("--spot_max_price can only be set when --priority is Spot")
        if len(str(namespace.spot_max_price).split(".")) > 1 and len(str(namespace.spot_max_price).split(".")[1]) > 5:
            raise CLIError("--spot_max_price can only include up to 5 decimal places")
        if namespace.spot_max_price <= 0 and not isclose(namespace.spot_max_price, -1.0, rel_tol=1e-06):
            raise CLIError(
                "--spot_max_price can only be any decimal value greater than zero, or -1 which indicates "
                "default price to be up-to on-demand")


def validate_acr(namespace):
    if namespace.attach_acr and namespace.detach_acr:
        raise CLIError('Cannot specify "--attach-acr" and "--detach-acr" at the same time.')
    if namespace.assignee_principal_type and not namespace.attach_acr:
        raise CLIError('The "--assignee-principal-type" argument can only be used with "--attach-acr".')
    if namespace.attach_acr:
        # Validate assignee_principal_type if specified
        if namespace.assignee_principal_type:
            valid_types = ['User', 'ServicePrincipal', 'Group']
            if namespace.assignee_principal_type not in valid_types:
                raise CLIError(
                    f"Invalid value for --assignee_principal_type. "
                    f"Allowed values are: {', '.join(valid_types)}"
                )


def validate_nodepool_tags(ns):
    """ Extracts multiple space-separated tags in key[=value] format """
    if isinstance(ns.nodepool_tags, list):
        tags_dict = {}
        for item in ns.nodepool_tags:
            tags_dict.update(validate_tag(item))
        ns.nodepool_tags = tags_dict


def validate_vnet_subnet_id(namespace):
    _validate_subnet_id(namespace.vnet_subnet_id, "--vnet-subnet-id")


def validate_pod_subnet_id(namespace):
    _validate_subnet_id(namespace.pod_subnet_id, "--pod-subnet-id")


def validate_pod_ip_allocation_mode(namespace):
    """Validates the pod ip allocation mode string."""
    if namespace.pod_ip_allocation_mode is not None:
        if namespace.pod_ip_allocation_mode not in (
            CONST_NETWORK_POD_IP_ALLOCATION_MODE_DYNAMIC_INDIVIDUAL,
            CONST_NETWORK_POD_IP_ALLOCATION_MODE_STATIC_BLOCK,
        ):
            raise InvalidArgumentValueError("--pod-ip-allocation-mode can only be DynamicIndividual or StaticBlock")


def validate_apiserver_subnet_id(namespace):
    _validate_subnet_id(namespace.apiserver_subnet_id, "--apiserver-subnet-id")


def validate_system_node_subnet_id(namespace):
    _validate_subnet_id(namespace.system_node_subnet_id, "--system-node-subnet-id")


def validate_node_subnet_id(namespace):
    _validate_subnet_id(namespace.node_subnet_id, "--node-subnet-id")


def _validate_subnet_id(subnet_id, name):
    if subnet_id is None or subnet_id == '':
        return
    from azure.mgmt.core.tools import is_valid_resource_id
    if not is_valid_resource_id(subnet_id):
        raise InvalidArgumentValueError(name + " is not a valid Azure resource ID.")


def validate_ppg(namespace):
    if namespace.ppg is not None:
        if namespace.ppg == '':
            return
        from azure.mgmt.core.tools import is_valid_resource_id
        if not is_valid_resource_id(namespace.ppg):
            raise CLIError("--ppg is not a valid Azure resource ID.")


def validate_node_public_ip_tags(ns):
    if isinstance(ns.node_public_ip_tags, list):
        tags_dict = {}
        for item in ns.node_public_ip_tags:
            tags_dict.update(validate_tag(item))
        ns.node_public_ip_tags = tags_dict


def validate_nodepool_labels(namespace):
    """Validates that provided node labels is a valid format"""

    if hasattr(namespace, 'nodepool_labels'):
        labels = namespace.nodepool_labels
    else:
        labels = namespace.labels

    if labels is None:
        return

    if isinstance(labels, list):
        labels_dict = {}
        for item in labels:
            labels_dict.update(validate_label(item))
        after_validation_labels = labels_dict
    else:
        after_validation_labels = validate_label(labels)

    if hasattr(namespace, 'nodepool_labels'):
        namespace.nodepool_labels = after_validation_labels
    else:
        namespace.labels = after_validation_labels


def validate_label(label):
    """Validates that provided label is a valid format"""
    prefix_regex = re.compile(r"^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$")
    name_regex = re.compile(r"^([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9]$")
    value_regex = re.compile(r"^(([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9])?$")

    if label == "":
        return {}
    kv = label.split('=')
    if len(kv) != 2:
        raise CLIError("Invalid label: %s. Label definition must be of format name=value." % label)
    name_parts = kv[0].split('/')
    if len(name_parts) == 1:
        name = name_parts[0]
    elif len(name_parts) == 2:
        prefix = name_parts[0]
        if not prefix or len(prefix) > 253:
            raise CLIError("Invalid label: %s. Label prefix can't be empty or more than 253 chars." % label)
        if not prefix_regex.match(prefix):
            raise CLIError("Invalid label: %s. Prefix part a DNS-1123 label must consist of lower case alphanumeric "
                           "characters or '-', and must start and end with an alphanumeric character" % label)
        name = name_parts[1]
    else:
        raise CLIError("Invalid label: %s. A qualified name must consist of alphanumeric characters, '-', '_' "
                       "or '.', and must start and end with an alphanumeric character (e.g. 'MyName',  or "
                       "'my.name',  or '123-abc') with an optional DNS subdomain prefix and '/' "
                       "(e.g. 'example.com/MyName')" % label)

    # validate label name
    if not name or len(name) > 63:
        raise CLIError("Invalid label: %s. Label name can't be empty or more than 63 chars." % label)
    if not name_regex.match(name):
        raise CLIError("Invalid label: %s. A qualified name must consist of alphanumeric characters, '-', '_' "
                       "or '.', and must start and end with an alphanumeric character (e.g. 'MyName',  or "
                       "'my.name',  or '123-abc') with an optional DNS subdomain prefix and '/' (e.g. "
                       "'example.com/MyName')" % label)

    # validate label value
    if len(kv[1]) > 63:
        raise CLIError("Invalid label: %s. Label must not be more than 63 chars." % label)
    if not value_regex.match(kv[1]):
        raise CLIError("Invalid label: %s. A valid label must be an empty string or consist of alphanumeric "
                       "characters, '-', '_' or '.', and must start and end with an alphanumeric character" % label)

    return {kv[0]: kv[1]}


def validate_max_surge(namespace):
    """validates parameters like max surge are postive integers or percents. less strict than RP"""
    if namespace.max_surge is None:
        return
    int_or_percent = namespace.max_surge
    if int_or_percent.endswith('%'):
        int_or_percent = int_or_percent.rstrip('%')

    try:
        if int(int_or_percent) < 0:
            raise CLIError("--max-surge must be positive")
    except ValueError:
        raise CLIError("--max-surge should be an int or percentage")


def validate_max_unavailable(namespace):
    """validates parameters max unavailable are positive integers or percents."""
    if namespace.max_unavailable is None:
        return
    int_or_percent = namespace.max_unavailable
    if int_or_percent.endswith('%'):
        int_or_percent = int_or_percent.rstrip('%')

    try:
        if int(int_or_percent) < 0:
            raise InvalidArgumentValueError("--max-unavailable must be positive")
    except ValueError:
        raise InvalidArgumentValueError("--max-unavailable should be an int or percentage")


def validate_assign_identity(namespace):
    if namespace.assign_identity is not None:
        if namespace.assign_identity == '':
            return
        from azure.mgmt.core.tools import is_valid_resource_id
        if not is_valid_resource_id(namespace.assign_identity):
            raise InvalidArgumentValueError("--assign-identity is not a valid Azure resource ID.")


def validate_assign_kubelet_identity(namespace):
    if namespace.assign_kubelet_identity is not None:
        if namespace.assign_kubelet_identity == '':
            return
        from azure.mgmt.core.tools import is_valid_resource_id
        if not is_valid_resource_id(namespace.assign_kubelet_identity):
            raise InvalidArgumentValueError("--assign-kubelet-identity is not a valid Azure resource ID.")


def validate_nodepool_id(namespace):
    from azure.mgmt.core.tools import is_valid_resource_id
    if not is_valid_resource_id(namespace.nodepool_id):
        raise InvalidArgumentValueError("--nodepool-id is not a valid Azure resource ID.")


def validate_snapshot_id(namespace):
    if namespace.snapshot_id:
        from azure.mgmt.core.tools import is_valid_resource_id
        if not is_valid_resource_id(namespace.snapshot_id):
            raise InvalidArgumentValueError("--snapshot-id is not a valid Azure resource ID.")


def validate_host_group_id(namespace):
    if namespace.host_group_id:
        from azure.mgmt.core.tools import is_valid_resource_id
        if not is_valid_resource_id(namespace.host_group_id):
            raise InvalidArgumentValueError("--host-group-id is not a valid Azure resource ID.")


def validate_crg_id(namespace):
    if namespace.crg_id is None:
        return
    from azure.mgmt.core.tools import is_valid_resource_id
    if not is_valid_resource_id(namespace.crg_id):
        raise InvalidArgumentValueError(
            "--crg-id is not a valid Azure resource ID.")


def extract_comma_separated_string(
    raw_string,
    enable_strip=False,
    extract_kv=False,
    allow_empty_value=False,
    keep_none=False,
    default_value=None,
    allow_appending_values_to_same_key=False,
):
    """Extract comma-separated string.

    If enable_strip is specified, will remove leading and trailing whitespace before each operation on the string.
    If extract_kv is specified, will extract key value pairs from the string with "=" as the delimiter and this would
    return a dictionary, otherwise keep the entire string.
    Option allow_empty_value is valid since extract_kv is specified. When the number of string segments split by "="
    is 1, the first segment is retained as the key and empty string would be set as its corresponding value without
    raising an exception.
    Option allow_appending_values_to_same_key is valid since extract_kv is specified. For the same key, the new value
    is appended to the existing value separated by commas.
    If keep_none is specified, will return None when input is None. Otherwise will return default_value if input is
    None or empty string.
    """
    if raw_string is None:
        if keep_none:
            return None
        return default_value
    if enable_strip:
        raw_string = raw_string.strip()
    if raw_string == "":
        return default_value

    result = {} if extract_kv else []
    for item in raw_string.split(","):
        if enable_strip:
            item = item.strip()
        if extract_kv:
            kv_list = item.split("=")
            if len(kv_list) in [1, 2]:
                key = kv_list[0]
                value = ""
                if len(kv_list) == 2:
                    value = kv_list[1]
                if not allow_empty_value and (value == "" or value.isspace()):
                    raise InvalidArgumentValueError(
                        "Empty value not allowed. The value '{}' of key '{}' in '{}' is empty. Raw input '{}'.".format(
                            value, key, item, raw_string
                        )
                    )
                if enable_strip:
                    key = key.strip()
                    value = value.strip()
                if allow_appending_values_to_same_key and key in result:
                    value = "{},{}".format(result[key], value)
                result[key] = value
            else:
                raise InvalidArgumentValueError(
                    "The format of '{}' in '{}' is incorrect, correct format should be "
                    "'Key1=Value1,Key2=Value2'.".format(
                        item, raw_string
                    )
                )
        else:
            result.append(item)
    return result


def validate_credential_format(namespace):
    if namespace.credential_format and \
        namespace.credential_format.lower() != "azure" and \
            namespace.credential_format.lower() != "exec":
        raise InvalidArgumentValueError("--format can only be azure or exec.")


def validate_keyvault_secrets_provider_disable_and_enable_parameters(namespace):
    if namespace.disable_secret_rotation and namespace.enable_secret_rotation:
        raise MutuallyExclusiveArgumentError(
            "Providing both --disable-secret-rotation and --enable-secret-rotation flags is invalid"
        )


def validate_defender_config_parameter(namespace):
    if namespace.defender_config and not namespace.enable_defender:
        raise RequiredArgumentMissingError("Please specify --enable-defnder")


def validate_defender_disable_and_enable_parameters(namespace):
    if namespace.disable_defender and namespace.enable_defender:
        raise ArgumentUsageError('Providing both --disable-defender and --enable-defender flags is invalid')


def validate_azure_keyvault_kms_key_id(namespace):
    key_id = namespace.azure_keyvault_kms_key_id
    if key_id:
        # pylint:disable=line-too-long
        err_msg = '--azure-keyvault-kms-key-id is not a valid Key Vault key ID. See https://learn.microsoft.com/en-us/azure/key-vault/general/about-keys-secrets-certificates#vault-name-and-object-name'

        https_prefix = "https://"
        if not key_id.startswith(https_prefix):
            raise InvalidArgumentValueError(err_msg)

        segments = key_id[len(https_prefix):].split("/")
        if len(segments) != 4 or segments[1] != "keys":
            raise InvalidArgumentValueError(err_msg)


def validate_azure_keyvault_kms_key_vault_resource_id(namespace):
    key_vault_resource_id = namespace.azure_keyvault_kms_key_vault_resource_id
    if key_vault_resource_id is None or key_vault_resource_id == '':
        return
    from azure.mgmt.core.tools import is_valid_resource_id
    if not is_valid_resource_id(key_vault_resource_id):
        raise InvalidArgumentValueError("--azure-keyvault-kms-key-vault-resource-id is not a valid Azure resource ID.")


def validate_image_cleaner_enable_disable_mutually_exclusive(namespace):
    enable_image_cleaner = namespace.enable_image_cleaner
    disable_image_cleaner = namespace.disable_image_cleaner

    if enable_image_cleaner and disable_image_cleaner:
        raise MutuallyExclusiveArgumentError(
            "Cannot specify --enable-image-cleaner and --disable-image-cleaner at the same time."
        )


def validate_registry_name(cmd, namespace):
    """Append login server endpoint suffix."""
    registry = namespace.acr
    suffixes = cmd.cli_ctx.cloud.suffixes
    # Some clouds do not define 'acr_login_server_endpoint' (e.g. AzureGermanCloud)
    from azure.cli.core.cloud import CloudSuffixNotSetException
    try:
        acr_suffix = suffixes.acr_login_server_endpoint
    except CloudSuffixNotSetException:
        acr_suffix = None
    if registry and acr_suffix:
        pos = registry.find(acr_suffix)
        if pos == -1:
            logger.warning("The login server endpoint suffix '%s' is automatically appended.", acr_suffix)
            namespace.acr = registry + acr_suffix


def sanitize_resource_id(resource_id):
    resource_id = resource_id.strip()
    if not resource_id.startswith("/"):
        resource_id = "/" + resource_id
    if resource_id.endswith("/"):
        resource_id = resource_id.rstrip("/")
    return resource_id.lower()


# pylint:disable=line-too-long
def validate_azuremonitor_privatelinkscope_resourceid(namespace):
    resource_id = namespace.ampls_resource_id
    if resource_id is None:
        return
    resource_id = sanitize_resource_id(resource_id)
    if (bool(re.match(r'/subscriptions/.*/resourcegroups/.*/providers/microsoft.insights/privatelinkscopes/.*', resource_id))) is False:
        raise InvalidArgumentValueError("--ampls-resource-id  not in the correct format. It should match `/subscriptions/<subscriptionId>/resourceGroups/<resourceGroupName>/providers/microsoft.insights/privatelinkscopes/<resourceName>`")


# pylint:disable=line-too-long
def validate_azuremonitorworkspaceresourceid(namespace):
    resource_id = namespace.azure_monitor_workspace_resource_id
    if resource_id is None:
        return
    resource_id = sanitize_resource_id(resource_id)
    if (bool(re.match(r'/subscriptions/.*/resourcegroups/.*/providers/microsoft.monitor/accounts/.*', resource_id))) is False:
        raise InvalidArgumentValueError("--azure-monitor-workspace-resource-id not in the correct format. It should match `/subscriptions/<subscriptionId>/resourceGroups/<resourceGroupName>/providers/microsoft.monitor/accounts/<resourceName>`")


# pylint:disable=line-too-long
def validate_grafanaresourceid(namespace):
    resource_id = namespace.grafana_resource_id
    if resource_id is None:
        return
    resource_id = sanitize_resource_id(resource_id)
    if (bool(re.match(r'/subscriptions/.*/resourcegroups/.*/providers/microsoft.dashboard/grafana/.*', resource_id))) is False:
        raise InvalidArgumentValueError("--grafana-resource-id not in the correct format. It should match `/subscriptions/<subscriptionId>/resourceGroups/<resourceGroupName>/providers/microsoft.dashboard/grafana/<resourceName>`")


def validate_os_sku(namespace):
    os_sku = namespace.os_sku
    if os_sku in [CONST_OS_SKU_MARINER, CONST_OS_SKU_CBLMARINER]:
        logger.warning(
            'The osSKU "%s" should be used going forward instead of "%s" or "%s". '
            'The osSKUs "%s" and "%s" will eventually be deprecated.',
            CONST_OS_SKU_AZURELINUX,
            CONST_OS_SKU_CBLMARINER,
            CONST_OS_SKU_MARINER,
            CONST_OS_SKU_CBLMARINER,
            CONST_OS_SKU_MARINER,
        )


def validate_utc_offset(namespace):
    """Validates --utc-offset for aks maintenanceconfiguration add/update commands."""
    if namespace.utc_offset is None:
        return
    # The regex here is used to match strings to the "+HH:MM" or "-HH:MM" time format.
    # The complete regex breakdown is as follows:
    # ^ matches the start of the string and $ matches to the end of the string, respectively.
    # [+-] match to either + or -
    # \d{2}:\d{2} will match to two digits followed by a colon and two digits. (example: +05:30 which is 5 hours and 30 minutes ahead or -12:00 which is 12 hours behind)
    utc_offset_regex = re.compile(r'^[+-]\d{2}:\d{2}$')
    found = utc_offset_regex.findall(namespace.utc_offset)
    if not found:
        raise InvalidArgumentValueError('--utc-offset must be in format: "+/-HH:mm". For example, "+05:30" and "-12:00".')


def validate_start_date(namespace):
    """Validates --start-date for aks maintenanceconfiguration add/update commands."""
    if namespace.start_date is None:
        return
    # The regex here is used to match strings to the "yyyy-MM-dd" date format.
    # The complete regex breakdown is as follows:
    # ^ matches the start of the string and $ matches to the end of the string, respectively.
    # ^\d{4}-\d{2}-\d{2} will match four digits, followed by a -, two digits, followed by a -. (example: 2023-01-01 which is January 1st 2023)
    start_dt_regex = re.compile(r'^\d{4}-\d{2}-\d{2}$')
    found = start_dt_regex.findall(namespace.start_date)
    if not found:
        raise InvalidArgumentValueError('--start-date must be in format: "yyyy-MM-dd". For example, "2023-01-01".')


def validate_start_time(namespace):
    """Validates --start-time for aks maintenanceconfiguration add/update commands."""
    if namespace.start_time is None:
        return
    # The regex here is used to match strings to the "HH:MM" or "HH:MM" time format.
    # The complete regex breakdown is as follows:
    # ^ matches the start of the string and $ matches to the end of the string, respectively.
    # \d{2}:\d{2} will match to two digits followed by a colon and two digits. (example: 09:30 which is 9 hours and 30 minutes or 17:00 which is 17 hours and 00 minutes)
    start_time_regex = re.compile(r'^\d{2}:\d{2}$')
    found = start_time_regex.findall(namespace.start_time)

    if not found:
        raise InvalidArgumentValueError('--start-time must be in format "HH:mm". For example, "09:30" and "17:00".')


def validate_force_upgrade_disable_and_enable_parameters(namespace):
    if namespace.disable_force_upgrade and namespace.enable_force_upgrade:
        raise MutuallyExclusiveArgumentError('Providing both --disable-force-upgrade and --enable-force-upgrade flags is invalid')


def validate_allowed_host_ports(namespace):
    if hasattr(namespace, "nodepool_allowed_host_ports"):
        host_ports = namespace.nodepool_allowed_host_ports
    else:
        host_ports = namespace.allowed_host_ports
    if not host_ports:
        return

    # Parse the port range. The format is either `<int>/<protocol>` or `<int>-<int>/<protocol>`.
    # e.g. `80/tcp` | `22/udp` | `4000-5000/tcp`
    regex = re.compile(r'^((\d+)|(\d+-\d+))/(tcp|udp)$')
    for port_range in host_ports:
        found = regex.findall(port_range.lower())
        if found:
            continue
        raise InvalidArgumentValueError(
            "--allowed-host-ports must be a space-separated list of port ranges in the format of <port-range>/<protocol>: '{}'".format(port_range)
        )


def validate_application_security_groups(namespace):
    is_nodepool_operation = False
    if hasattr((namespace), "nodepool_asg_ids"):
        is_nodepool_operation = True
        asg_ids = namespace.nodepool_asg_ids
        host_ports = namespace.nodepool_allowed_host_ports
    else:
        asg_ids = namespace.asg_ids
        host_ports = namespace.allowed_host_ports

    if not asg_ids:
        return

    if not host_ports:
        if is_nodepool_operation:
            raise ArgumentUsageError(
                '--nodepool-asg-ids must be used with --nodepool-allowed-host-ports'
            )
        raise ArgumentUsageError(
            '--asg-ids must be used with --allowed-host-ports'
        )

    from azure.mgmt.core.tools import is_valid_resource_id
    for asg in asg_ids:
        if not is_valid_resource_id(asg):
            raise InvalidArgumentValueError(asg + " is not a valid Azure resource ID.")


def validate_azure_service_mesh_revision(namespace):
    """Validates the user provided revision parameter for azure service mesh commands."""
    if namespace.revision is None:
        return
    revision = namespace.revision
    asm_revision_regex = re.compile(r'^asm-\d+-\d+$')
    found = asm_revision_regex.findall(revision)
    if not found:
        raise InvalidArgumentValueError(f"Revision {revision} is not supported by the service mesh add-on.")


def validate_disable_windows_outbound_nat(namespace):
    """Validates disable_windows_outbound_nat can only be used on Windows."""
    if namespace.disable_windows_outbound_nat:
        if hasattr(namespace, 'os_type') and str(namespace.os_type).lower() != "windows":
            raise ArgumentUsageError(
                '--disable-windows-outbound-nat can only be set for Windows nodepools')


def validate_message_of_the_day(namespace):
    """Validates message of the day can only be used on Linux."""
    if namespace.message_of_the_day is not None and namespace.message_of_the_day != "":
        if namespace.os_type is not None and namespace.os_type != "Linux":
            raise ArgumentUsageError(
                '--message-of-the-day can only be set for linux nodepools')


def validate_bootstrap_container_registry_resource_id(namespace):
    container_registry_resource_id = namespace.bootstrap_container_registry_resource_id
    if container_registry_resource_id is None or container_registry_resource_id == '':
        return
    from azure.mgmt.core.tools import is_valid_resource_id
    if not is_valid_resource_id(container_registry_resource_id):
        raise InvalidArgumentValueError("--bootstrap-container-registry-resource-id is not a valid Azure resource ID.")


def validate_custom_ca_trust_certificates(namespace):
    """Validates Custom CA Trust Certificates can only be used on Linux."""
    if namespace.custom_ca_trust_certificates is not None and namespace.custom_ca_trust_certificates != "":
        if hasattr(namespace, 'os_type') and namespace.os_type != "Linux":
            raise ArgumentUsageError(
                '--custom-ca-trust-certificates can only be set for linux nodepools')


def validate_gateway_prefix_size(namespace):
    """Validates the gateway prefix size."""
    if namespace.gateway_prefix_size is not None:
        if not hasattr(namespace, 'mode') or namespace.mode != CONST_NODEPOOL_MODE_GATEWAY:
            raise ArgumentUsageError("--gateway-prefix-size can only be set for Gateway-mode nodepools")
        if namespace.gateway_prefix_size < 28 or namespace.gateway_prefix_size > 31:
            raise InvalidArgumentValueError("--gateway-prefix-size must be in the range [28, 31]")


def validate_artifact_streaming(namespace):
    """Validates artifact streaming flags for mutual exclusivity and OS support."""
    enable_artifact_streaming = getattr(namespace, "enable_artifact_streaming", False)
    disable_artifact_streaming = getattr(namespace, "disable_artifact_streaming", False)

    if enable_artifact_streaming and disable_artifact_streaming:
        raise MutuallyExclusiveArgumentError(
            "Cannot specify both --enable-artifact-streaming and --disable-artifact-streaming at the same time."
        )

    if hasattr(namespace, "os_type") and str(namespace.os_type).lower() == "windows":
        if enable_artifact_streaming:
            raise ArgumentUsageError('--enable-artifact-streaming can only be set for Linux nodepools')
        if disable_artifact_streaming:
            raise ArgumentUsageError('--disable-artifact-streaming can only be set for Linux nodepools')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from collections import OrderedDict

from azure.cli.core.commands import CliCommandType
from azure.cli.core.profiles import ResourceType

from azure.cli.command_modules.keyvault._client_factory import (
    get_client, get_client_factory, Clients, data_plane_azure_keyvault_ekm_client)

from azure.cli.command_modules.keyvault._transformers import (
    filter_out_managed_resources,
    multi_transformers, transform_key_decryption_output, keep_max_results, transform_key_list_output,
    transform_key_output, transform_key_encryption_output, transform_key_random_output,
    transform_secret_list, transform_deleted_secret_list, transform_secret_set,
    transform_secret_set_attributes, transform_secret_show_deleted, transform_secret_delete, transform_secret_recover,
    transform_certificate_create, transform_certificate_list, transform_certificate_list_deleted,
    transform_certificate_show, transform_certificate_show_deleted, transform_certificate_delete,
    transform_certificate_recover, transform_certificate_contact_list, transform_certificate_contact_add,
    transform_certificate_issuer_create, transform_certificate_issuer_list, transform_certificate_issuer_admin_list)

from azure.cli.command_modules.keyvault._format import transform_secret_list_table

from azure.cli.command_modules.keyvault._validators import (
    process_secret_set_namespace, validate_key_create,
    validate_private_endpoint_connection_id, validate_role_assignment_args,
    validate_ekm_connection_base, validate_ekm_connection_create, validate_ekm_connection_update)


def transform_assignment_list(result):
    return [OrderedDict([('Principal', r['principalName']),
                         ('RoleName', r['roleName']),
                         ('Scope', r['scope'])]) for r in result]


def transform_definition_list(result):
    return [OrderedDict([('Name', r['name']), ('RoleName', r['roleName']), ('Type', r['type']),
                         ('Description', r['description'])]) for r in result]


# pylint: disable=too-many-locals, too-many-statements
def load_command_table(self, _):
    # region Command Types
    mgmt_vaults_entity = get_client(ResourceType.MGMT_KEYVAULT, Clients.vaults)
    mgmt_pec_entity = get_client(ResourceType.MGMT_KEYVAULT, Clients.private_endpoint_connections)
    mgmt_plr_entity = get_client(ResourceType.MGMT_KEYVAULT, Clients.private_link_resources)
    data_key_entity = get_client(ResourceType.DATA_KEYVAULT_KEYS)
    data_certificate_entity = get_client(ResourceType.DATA_KEYVAULT_CERTIFICATES)
    data_secret_entity = get_client(ResourceType.DATA_KEYVAULT_SECRETS)

    mgmt_hsms_entity = get_client(ResourceType.MGMT_KEYVAULT, Clients.managed_hsms)
    mgmt_hsms_regions_entity = get_client(ResourceType.MGMT_KEYVAULT, Clients.mhsm_regions)
    data_security_domain_entity = get_client(ResourceType.DATA_KEYVAULT_SECURITY_DOMAIN)
    data_backup_entity = get_client(ResourceType.DATA_KEYVAULT_ADMINISTRATION_BACKUP)
    data_access_control_entity = get_client(ResourceType.DATA_KEYVAULT_ADMINISTRATION_ACCESS_CONTROL)
    data_setting_entity = get_client(ResourceType.DATA_KEYVAULT_ADMINISTRATION_SETTING)

    kv_vaults_custom = CliCommandType(
        operations_tmpl='azure.cli.command_modules.keyvault.custom#{}',
        client_factory=get_client_factory(ResourceType.MGMT_KEYVAULT, Clients.vaults)
    )
    kv_hsms_custom = CliCommandType(
        operations_tmpl='azure.cli.command_modules.keyvault.custom#{}',
        client_factory=get_client_factory(ResourceType.MGMT_KEYVAULT, Clients.managed_hsms)
    )

    data_ekm_custom = CliCommandType(
        operations_tmpl='azure.cli.command_modules.keyvault.custom#{}',
        client_factory=data_plane_azure_keyvault_ekm_client
    )
    # endregion

    # Management Plane Commands
    with self.command_group('keyvault', mgmt_vaults_entity.command_type,
                            client_factory=mgmt_vaults_entity.client_factory) as g:
        g.custom_command('create', 'create_vault_or_hsm', supports_no_wait=True)
        g.custom_command('recover', 'recover_vault_or_hsm', supports_no_wait=True)
        g.custom_command('list', 'list_vault_or_hsm')
        g.custom_show_command('show', 'get_vault_or_hsm')
        g.custom_command('delete', 'delete_vault_or_hsm', supports_no_wait=True)
        g.custom_command('purge', 'purge_vault_or_hsm', supports_no_wait=True)
        g.custom_command('set-policy', 'set_policy', supports_no_wait=True)
        g.custom_command('delete-policy', 'delete_policy', supports_no_wait=True)
        g.custom_command('list-deleted', 'list_deleted_vault_or_hsm')
        g.custom_command('show-deleted', 'get_deleted_vault_or_hsm')
        g.generic_update_command(
            'update', setter_name='update_vault_setter', setter_type=kv_vaults_custom,
            custom_func_name='update_vault',
            supports_no_wait=True)
        g.wait_command('wait')

    with self.command_group('keyvault', mgmt_hsms_entity.command_type,
                            client_factory=mgmt_hsms_entity.client_factory) as g:
        g.generic_update_command(
            'update-hsm', setter_name='update_hsm_setter', setter_type=kv_hsms_custom,
            custom_func_name='update_hsm', supports_no_wait=True)
        g.custom_wait_command('wait-hsm', 'wait_hsm')
        g.custom_command('check-name', 'check_name_availability')

    with self.command_group('keyvault network-rule',
                            mgmt_vaults_entity.command_type,
                            client_factory=mgmt_vaults_entity.client_factory) as g:
        g.custom_command('add', 'add_network_rule_for_vault_or_hsm', supports_no_wait=True)
        g.custom_command('remove', 'remove_network_rule_for_vault_or_hsm', supports_no_wait=True)
        g.custom_command('list', 'list_network_rules')
        g.custom_wait_command('wait', 'wait_vault_or_hsm')

    with self.command_group('keyvault private-endpoint-connection',
                            mgmt_pec_entity.command_type,
                            client_factory=mgmt_pec_entity.client_factory) as g:
        g.custom_command('approve', 'approve_private_endpoint_connection', supports_no_wait=True,
                         validator=validate_private_endpoint_connection_id)
        g.custom_command('reject', 'reject_private_endpoint_connection', supports_no_wait=True,
                         validator=validate_private_endpoint_connection_id)
        g.custom_command('delete', 'delete_private_endpoint_connection',
                         validator=validate_private_endpoint_connection_id, supports_no_wait=True)
        g.custom_command('list', 'list_private_endpoint_connection')
        g.custom_show_command('show', 'show_private_endpoint_connection',
                              validator=validate_private_endpoint_connection_id)
        g.custom_wait_command('wait', 'show_private_endpoint_connection',
                              validator=validate_private_endpoint_connection_id)

    with self.command_group('keyvault private-link-resource',
                            mgmt_plr_entity.command_type,
                            client_factory=mgmt_plr_entity.client_factory) as g:
        from azure.cli.core.commands.transform import gen_dict_to_list_transform
        g.custom_command('list', 'list_private_link_resource', transform=gen_dict_to_list_transform(key='value'))

    # Data Plane Commands
    with self.command_group('keyvault backup', data_backup_entity.command_type) as g:
        g.keyvault_custom('start', 'full_backup')

    with self.command_group('keyvault restore', data_backup_entity.command_type) as g:
        g.keyvault_custom('start', 'full_restore')

    with self.command_group('keyvault security-domain', data_security_domain_entity.command_type) as g:
        g.keyvault_custom('init-recovery', 'security_domain_init_recovery')
        g.keyvault_custom('restore-blob', 'security_domain_restore_blob')
        g.keyvault_custom('upload', 'security_domain_upload', supports_no_wait=True)
        g.keyvault_custom('download', 'security_domain_download', supports_no_wait=True)
        g.keyvault_custom('wait', '_wait_security_domain_operation')

    with self.command_group('keyvault ekm-connection', command_type=data_ekm_custom, is_preview=True) as g:
        g.keyvault_custom('create', 'create_ekm_connection', validator=validate_ekm_connection_create)
        g.keyvault_custom('update', 'update_ekm_connection', validator=validate_ekm_connection_update)
        g.keyvault_custom('show', 'get_ekm_connection', validator=validate_ekm_connection_base)
        g.keyvault_custom('check', 'check_ekm_connection', validator=validate_ekm_connection_base)
        g.keyvault_custom('delete', 'delete_ekm_connection', validator=validate_ekm_connection_base)

    with self.command_group('keyvault ekm-connection certificate', command_type=data_ekm_custom, is_preview=True) as g:
        g.keyvault_custom('show', 'get_ekm_certificate', validator=validate_ekm_connection_base)

    with self.command_group('keyvault key', data_key_entity.command_type) as g:
        g.keyvault_custom('create', 'create_key', transform=transform_key_output, validator=validate_key_create)
        g.keyvault_command('set-attributes', 'update_key_properties', transform=transform_key_output)
        g.keyvault_command('show', 'get_key', transform=transform_key_output)
        g.keyvault_custom('get-attestation', 'get_key_attestation')
        g.keyvault_custom('import', 'import_key', transform=transform_key_output)
        g.keyvault_custom('get-policy-template', 'get_policy_template', is_preview=True)
        g.keyvault_custom('encrypt', 'encrypt_key', is_preview=True, transform=transform_key_encryption_output)
        g.keyvault_custom('decrypt', 'decrypt_key', is_preview=True, transform=transform_key_decryption_output)
        g.keyvault_custom('list', 'list_keys',
                          transform=multi_transformers(keep_max_results, transform_key_list_output))
        g.keyvault_command('list-versions', 'list_properties_of_key_versions',
                           transform=multi_transformers(keep_max_results, transform_key_list_output))
        g.keyvault_command('list-deleted', 'list_deleted_keys',
                           transform=multi_transformers(keep_max_results, transform_key_list_output))
        g.keyvault_command('show-deleted', 'get_deleted_key', transform=transform_key_output)
        g.keyvault_custom('delete', 'delete_key', transform=transform_key_output)
        g.keyvault_custom('recover', 'recover_key', transform=transform_key_output)
        g.keyvault_command('purge', 'purge_deleted_key')
        g.keyvault_custom('download', 'download_key')
        g.keyvault_custom('backup', 'backup_key')
        g.keyvault_custom('restore', 'restore_key', supports_no_wait=True, transform=transform_key_output)

    with self.command_group('keyvault key', data_key_entity.command_type) as g:
        g.keyvault_command('random', 'get_random_bytes', transform=transform_key_random_output)
        g.keyvault_command('rotate', 'rotate_key', transform=transform_key_output)
        g.keyvault_custom('sign', 'sign_key')
        g.keyvault_custom('verify', 'verify_key')

    with self.command_group('keyvault key rotation-policy', data_key_entity.command_type) as g:
        g.keyvault_command('show', 'get_key_rotation_policy', )
        g.keyvault_custom('update', 'update_key_rotation_policy')

    # secret track2
    with self.command_group('keyvault secret', data_secret_entity.command_type) as g:
        g.keyvault_custom('list', "list_secret",
                          transform=multi_transformers(
                              filter_out_managed_resources,
                              keep_max_results,
                              transform_secret_list),
                          table_transformer=transform_secret_list_table)
        g.keyvault_command('list-versions', 'list_properties_of_secret_versions',
                           transform=multi_transformers(
                               keep_max_results,
                               transform_secret_list))
        g.keyvault_command('list-deleted', 'list_deleted_secrets',
                           transform=multi_transformers(
                               keep_max_results,
                               transform_deleted_secret_list))
        g.keyvault_command('set', 'set_secret', validator=process_secret_set_namespace, transform=transform_secret_set)
        g.keyvault_command('set-attributes', 'update_secret_properties', transform=transform_secret_set_attributes)
        g.keyvault_command('show', 'get_secret', transform=transform_secret_set)
        g.keyvault_command('show-deleted', 'get_deleted_secret', transform=transform_secret_show_deleted)
        g.keyvault_command('delete', 'begin_delete_secret', transform=transform_secret_delete,
                           deprecate_info=g.deprecate(
                               tag_func=lambda x: '',
                               message_func=lambda x:
                               'Warning! If you have soft-delete protection enabled on this key vault, this secret '
                               'will be moved to the soft deleted state. You will not be able to create a secret '
                               'with the same name within this key vault until the secret has been purged from the '
                               'soft-deleted state. Please see the following documentation for additional guidance.'
                               '\nhttps://learn.microsoft.com/azure/key-vault/general/soft-delete-overview'))
        g.keyvault_command('purge', 'purge_deleted_secret')
        g.keyvault_command('recover', 'begin_recover_deleted_secret', transform=transform_secret_recover)
        g.keyvault_custom('download', 'download_secret')
        g.keyvault_custom('backup', 'backup_secret')
        g.keyvault_custom('restore', 'restore_secret', transform=transform_secret_set_attributes)

    # certificate track2
    with self.command_group('keyvault certificate', data_certificate_entity.command_type) as g:
        g.keyvault_custom('create', 'create_certificate', transform=transform_certificate_create)
        g.keyvault_command('list', 'list_properties_of_certificates',
                           transform=multi_transformers(
                               keep_max_results,
                               transform_certificate_list))
        g.keyvault_command('list-versions', 'list_properties_of_certificate_versions',
                           transform=multi_transformers(
                               keep_max_results,
                               transform_certificate_list))
        g.keyvault_command('list-deleted', 'list_deleted_certificates',
                           transform=multi_transformers(
                               keep_max_results,
                               transform_certificate_list_deleted))
        g.keyvault_command('show', 'get_certificate_version', transform=transform_certificate_show)
        g.keyvault_command('show-deleted', 'get_deleted_certificate', transform=transform_certificate_show_deleted)
        g.keyvault_command('delete', 'begin_delete_certificate', deprecate_info=g.deprecate(
            tag_func=lambda x: '',
            message_func=lambda x: 'Warning! If you have soft-delete protection enabled on this key vault, this '
                                   'certificate will be moved to the soft deleted state. You will not be able to '
                                   'create a certificate with the same name within this key vault until the '
                                   'certificate has been purged from the soft-deleted state. Please see the following '
                                   'documentation for additional guidance.\n'
                                   'https://learn.microsoft.com/azure/key-vault/general/soft-delete-overview'),
                           transform=transform_certificate_delete)
        g.keyvault_command('purge', 'purge_deleted_certificate')
        g.keyvault_command('recover', 'begin_recover_deleted_certificate', transform=transform_certificate_recover)
        g.keyvault_custom('set-attributes', 'set_attributes_certificate', transform=transform_certificate_show)
        g.keyvault_command('import', 'import_certificate', transform=transform_certificate_show)
        g.keyvault_custom('download', 'download_certificate')
        g.keyvault_custom('get-default-policy', 'get_default_policy')

    with self.command_group('keyvault certificate', data_certificate_entity.command_type) as g:
        g.keyvault_custom('backup', 'backup_certificate')
        g.keyvault_custom('restore', 'restore_certificate', transform=transform_certificate_show)

    with self.command_group('keyvault certificate pending', data_certificate_entity.command_type) as g:
        g.keyvault_command('merge', 'merge_certificate', transform=transform_certificate_show)
        g.keyvault_command('show', 'get_certificate_operation', transform=transform_certificate_create)
        g.keyvault_command('delete', 'delete_certificate_operation', transform=transform_certificate_create)

    with self.command_group('keyvault certificate contact', data_certificate_entity.command_type) as g:
        g.keyvault_command('list', 'get_contacts', transform=transform_certificate_contact_list)
        g.keyvault_custom('add', 'add_certificate_contact', transform=transform_certificate_contact_add)
        g.keyvault_custom('delete', 'delete_certificate_contact', transform=transform_certificate_contact_add)

    with self.command_group('keyvault certificate issuer', data_certificate_entity.command_type) as g:
        g.keyvault_custom('create', 'create_certificate_issuer', transform=transform_certificate_issuer_create)
        g.keyvault_custom('update', 'update_certificate_issuer', transform=transform_certificate_issuer_create)
        g.keyvault_command('list', 'list_properties_of_issuers',
                           transform=multi_transformers(
                               keep_max_results,
                               transform_certificate_issuer_list))
        g.keyvault_command('show', 'get_issuer', transform=transform_certificate_issuer_create)
        g.keyvault_command('delete', 'delete_issuer', transform=transform_certificate_issuer_create)

    with self.command_group('keyvault certificate issuer admin', data_certificate_entity.command_type) as g:
        g.keyvault_custom('add', 'add_certificate_issuer_admin')
        g.keyvault_command('list', 'get_issuer',
                           transform=multi_transformers(
                               keep_max_results,
                               transform_certificate_issuer_admin_list))
        g.keyvault_custom('delete', 'delete_certificate_issuer_admin')

    with self.command_group('keyvault role', data_access_control_entity.command_type):
        pass

    with self.command_group('keyvault role assignment', data_access_control_entity.command_type) as g:
        g.keyvault_custom('delete', 'delete_role_assignment', validator=validate_role_assignment_args)
        g.keyvault_custom('list', 'list_role_assignments', table_transformer=transform_assignment_list)
        g.keyvault_custom('create', 'create_role_assignment')

    with self.command_group('keyvault role definition', data_access_control_entity.command_type) as g:
        g.keyvault_custom('list', 'list_role_definitions', table_transformer=transform_definition_list)
        g.keyvault_custom('create', 'create_role_definition')
        g.keyvault_custom('update', 'update_role_definition')
        g.keyvault_custom('delete', 'delete_role_definition')
        g.keyvault_custom('show', 'show_role_definition')

    with self.command_group('keyvault setting', data_setting_entity.command_type) as g:
        g.keyvault_command('list', 'list_settings')
        g.keyvault_command('show', 'get_setting')
        g.keyvault_custom('update', 'update_hsm_setting')

    with self.command_group('keyvault region', mgmt_hsms_regions_entity.command_type,
                            client_factory=mgmt_hsms_regions_entity.client_factory) as g:
        g.command('list', 'list_by_resource', client_factory=mgmt_hsms_regions_entity.client_factory)

    with self.command_group('keyvault region', mgmt_hsms_entity.command_type,
                            client_factory=mgmt_hsms_entity.client_factory) as g:
        g.custom_command('add', 'add_hsm_region', supports_no_wait=True)
        g.custom_command('remove', 'remove_hsm_region', supports_no_wait=True)
        g.wait_command('wait')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from azure.cli.command_modules.acs.azuremonitormetrics.constants import DC_API
from azure.cli.command_modules.acs.azuremonitormetrics.dc.defaults import get_default_dcra_name
from knack.util import CLIError


# pylint: disable=line-too-long
def create_dcra(cmd, cluster_region, cluster_subscription, cluster_resource_group_name, cluster_name, dcr_resource_id):
    cluster_resource_id = "/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.ContainerService/managedClusters/{2}".format(
        cluster_subscription,
        cluster_resource_group_name,
        cluster_name
    )
    dcra_name = get_default_dcra_name()
    dcra_resource_id = "/subscriptions/{0}/resourceGroups/{1}/providers/Microsoft.Insights/dataCollectionRuleAssociations/{2}".format(
        cluster_subscription,
        cluster_resource_group_name,
        dcra_name
    )
    description_str = "Promtheus data collection association between DCR, DCE and target AKS resource"
    association_body = {
        "location": cluster_region,
        "properties": {
            "dataCollectionRuleId": dcr_resource_id,
            "description": description_str
        }
    }
    from azure.cli.command_modules.acs._client_factory import get_resources_client
    resources = get_resources_client(cmd.cli_ctx, cluster_subscription)
    try:
        resources.begin_create_or_update_by_id(
            f"{cluster_resource_id}/providers/Microsoft.Insights/dataCollectionRuleAssociations/{dcra_name}",
            DC_API,
            association_body
        )
        return dcra_resource_id
    except Exception as error:
        raise CLIError(error)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.testsdk import ScenarioTest, ResourceGroupPreparer


class TestAccountScenarios(ScenarioTest):
    @ResourceGroupPreparer(name_prefix='cli_test_monitor_account')
    def test_monitor_account(self, resource_group):
        self.kwargs.update({
            'account': self.create_random_name('ac', 10)
        })
        self.cmd('monitor account create -n {account} -g {rg}', checks=[
            self.check('name', '{account}'),
            self.check('resourceGroup', '{rg}')
        ])
        self.cmd('monitor account update -n {account} -g {rg} --tags {{tag:test,tag2:test2}}', checks=[
            self.check('name', '{account}'),
            self.check('resourceGroup', '{rg}'),
            self.check('tags.tag', 'test'),
            self.check('tags.tag2', 'test2')
        ])
        self.cmd('monitor account show -n {account} -g {rg}', checks=[
            self.check('name', '{account}'),
            self.check('resourceGroup', '{rg}'),
            self.check('tags.tag', 'test'),
            self.check('tags.tag2', 'test2')
        ])
        self.cmd('monitor account list -g {rg}', checks=[
            self.check('[0].name', '{account}'),
            self.check('[0].resourceGroup', '{rg}'),
            self.check('[0].tags.tag', 'test'),
            self.check('[0].tags.tag2', 'test2')
        ])



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest

from azure.keyvault.keys import DeletedKey, JsonWebKey, KeyProperties, KeyVaultKey

from azure.cli.command_modules.keyvault._transformers import (
    transform_key_list_output,
    transform_key_output,
)


class _Attrs:
    def __init__(self, key_size=None):
        self.enabled = True
        self.not_before = None
        self.expires = None
        self.created = None
        self.updated = None
        self.recovery_level = "Recoverable"
        self.recoverable_days = 7
        self.exportable = False
        self.hsm_platform = None
        self.key_size = key_size


def _kid(name):
    return "https://example.vault.azure.net/keys/{}/abc".format(name)


def _make_key_properties(name, key_size=None):
    return KeyProperties(key_id=_kid(name), attributes=_Attrs(key_size=key_size))


def _make_key(name, key_size=None, kty="oct"):
    key = KeyVaultKey.__new__(KeyVaultKey)
    key._properties = _make_key_properties(name, key_size=key_size)
    key._key_material = JsonWebKey(kty=kty, key_ops=["encrypt", "decrypt"])
    return key


def _make_deleted_key(name, key_size=None, kty="oct"):
    # DeletedKey.__init__ forwards to KeyVaultKey.__init__, which still
    # requires the positional key_id arg; bypass with __new__.
    deleted = DeletedKey.__new__(DeletedKey)
    deleted._properties = _make_key_properties(name, key_size=key_size)
    deleted._key_material = JsonWebKey(kty=kty)
    deleted._deleted_date = None
    deleted._recovery_id = "https://example.vault.azure.net/deletedkeys/{}".format(name)
    deleted._scheduled_purge_date = None
    return deleted


class KeyVaultTransformersTests(unittest.TestCase):

    def test_show_aes_key_surfaces_key_size(self):
        out = transform_key_output(_make_key("aes-256", key_size=256, kty="oct"))

        self.assertEqual(out["attributes"]["keySize"], 256)

    def test_show_aes_hsm_key_surfaces_key_size(self):
        out = transform_key_output(_make_key("aes-128", key_size=128, kty="oct-HSM"))

        self.assertEqual(out["attributes"]["keySize"], 128)

    def test_show_rsa_key_returns_none_key_size(self):
        out = transform_key_output(_make_key("rsa", key_size=None, kty="RSA"))

        self.assertIn("keySize", out["attributes"])
        self.assertIsNone(out["attributes"]["keySize"])

    def test_show_deleted_key_includes_key_size_and_deletion_fields(self):
        out = transform_key_output(_make_deleted_key("deleted-aes", key_size=256))

        self.assertEqual(out["attributes"]["keySize"], 256)
        self.assertIn("deletedDate", out)
        self.assertIn("scheduledPurgeDate", out)
        self.assertTrue(out["recoveryId"].endswith("/deletedkeys/deleted-aes"))

    def test_show_passes_through_unknown_types(self):
        sentinel = {"already": "transformed"}

        self.assertIs(transform_key_output(sentinel), sentinel)

    def test_list_surfaces_key_size_per_item(self):
        items = [
            _make_key_properties("aes-256", key_size=256),
            _make_key_properties("aes-128", key_size=128),
            _make_key_properties("rsa", key_size=None),
        ]

        out = transform_key_list_output(items)

        self.assertEqual([k["keySize"] for k in out], [256, 128, None])
        self.assertEqual([k["name"] for k in out], ["aes-256", "aes-128", "rsa"])

    def test_list_surfaces_key_size_for_deleted_entries(self):
        items = [
            _make_deleted_key("deleted-aes", key_size=256),
            _make_key_properties("live-aes", key_size=128),
        ]

        out = transform_key_list_output(items)

        self.assertEqual(out[0]["keySize"], 256)
        self.assertIn("deletedDate", out[0])
        self.assertIn("scheduledPurgeDate", out[0])
        self.assertEqual(out[1]["keySize"], 128)
        self.assertNotIn("deletedDate", out[1])

    def test_list_empty_returns_input(self):
        self.assertEqual(transform_key_list_output([]), [])
        self.assertIsNone(transform_key_list_output(None))

    def test_list_passes_through_non_key_properties(self):
        original = [{"not": "a key"}]

        self.assertIs(transform_key_list_output(original), original)


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Support for Entra Agentic Sessions.

When CLI runs inside an agent context (e.g., Copilot, Azure MCP), the orchestrator sets the
COPILOT_AGENT_SESSION_ID environment variable. CLI reads it and passes it to MSAL as both:
  - A query parameter (`client_session`) so ESTS can identify the agentic session
  - A claims challenge so ESTS embeds an agentic marker claim in the token (and MSAL bypasses
    the access token cache to ensure a fresh, agent-tagged token is always fetched)

This enables downstream systems (RBAC, Defender, Purview) to enforce differentiated policies
for agent-driven vs. human-driven operations.
"""

import json
import os

from knack.log import get_logger

logger = get_logger(__name__)

COPILOT_AGENT_SESSION_ID = "COPILOT_AGENT_SESSION_ID"


def build_agentic_session_params():
    """Read COPILOT_AGENT_SESSION_ID and build the agentic claims challenge.

    :returns: (session_id, claims_challenge) — both None when env var is not set.
    """
    session_id = os.environ.get(COPILOT_AGENT_SESSION_ID) or None
    if not session_id:
        return None, None

    logger.debug("Agentic session detected (COPILOT_AGENT_SESSION_ID is set)")

    claims_challenge = json.dumps({
        "access_token": {
            "xms_cli_sid": {"values": [session_id]}
        }
    })
    return session_id, claims_challenge


def merge_access_token_claims(existing_claims, new_claims):
    """Merge new claims into an existing claims_challenge JSON string.

    :param existing_claims: Existing claims_challenge JSON string (or None).
    :param new_claims: New claims_challenge JSON string to merge in. Must not be None or empty,
        and must contain a non-empty ``access_token`` object.
    :returns: Merged claims_challenge JSON string.
    :raises ValueError: If ``new_claims`` is None, empty, or does not contain a non-empty
        ``access_token`` object.
    """
    if not new_claims:
        raise ValueError("new_claims must not be None or empty")
    new_access_token = json.loads(new_claims).get("access_token")
    if not new_access_token:
        raise ValueError("new_claims must contain a non-empty access_token")

    claims_dict = json.loads(existing_claims) if existing_claims else {}
    claims_dict["access_token"] = claims_dict.get("access_token") or {}
    claims_dict["access_token"].update(new_access_token)
    return json.dumps(claims_dict)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "policy exemption list",
)
class List(AAZCommand):
    """Retrieve all applicable policy exemptions.

    Retrieve the list of all policy assignments applicable to the given subscription or management group.

    :example: List policy exemptions that apply to a management group
        az policy exemption list --management-group DevOrg --filter atScopeAndBelow()

    :example: List policy exemptions that apply to a resource group
        az policy exemption list --resource-group TestResourceGroup

    :example: List policy exemptions that apply to a subscription
        az policy exemption list
    """

    _aaz_info = {
        "version": "2026-01-01-preview",
        "resources": [
            ["mgmt-plane", "/providers/microsoft.management/managementgroups/{}/providers/microsoft.authorization/policyexemptions", "2026-01-01-preview"],
            ["mgmt-plane", "/subscriptions/{}/providers/microsoft.authorization/policyexemptions", "2026-01-01-preview"],
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.authorization/policyexemptions", "2026-01-01-preview"],
        ]
    }

    AZ_SUPPORT_PAGINATION = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_paging(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.management_group = AAZStrArg(
            options=["--management-group"],
            help={"short-summary": "The management group.", "long-summary": "Indicates that policy exemptions whose scope covers the management group with the given name should be listed."},
            fmt=AAZStrArgFormat(
                min_length=1,
            ),
        )
        _args_schema.resource_group = AAZResourceGroupNameArg()
        _args_schema.filter = AAZStrArg(
            options=["--filter"],
            help={"short-summary": "Filter list results.", "long-summary": "The filter to limit list results. Valid values are: 'atScope()', 'atExactScope()', 'atScopeAndBelow()' or 'policyDefinitionId eq '{value}''. If filter is not provided, no filtering is performed. If filter atScope() is provided, the returned list includes all policy assignments that apply to the given scope, which is everything in the unfiltered list except those applied to sub scopes contained within the given scope. If filter atExactScope() is provided, the returned list includes all policy assignments at the given scope.  If filter atScopeAndBelow() is provided, the returned list includes all policy assignments at the given scope and those in sub scopes contained within the given scope. If filter policyDefinitionId eq '{value}' is provided, the returned list includes all policy assignments of the policy definition whose id is {value}."},
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        condition_0 = has_value(self.ctx.args.management_group)
        condition_1 = has_value(self.ctx.subscription_id) and has_value(self.ctx.args.resource_group) is not True
        condition_2 = has_value(self.ctx.args.resource_group) and has_value(self.ctx.subscription_id)
        if condition_0:
            self.PolicyExemptionsListForManagementGroup(ctx=self.ctx)()
        if condition_1:
            self.PolicyExemptionsList(ctx=self.ctx)()
        if condition_2:
            self.PolicyExemptionsListForResourceGroup(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance.value, client_flatten=True)
        next_link = self.deserialize_output(self.ctx.vars.instance.next_link)
        return result, next_link

    class PolicyExemptionsListForManagementGroup(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/providers/Microsoft.Management/managementGroups/{managementGroupId}/providers/Microsoft.Authorization/policyExemptions",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "managementGroupId", self.ctx.args.management_group,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "$filter", self.ctx.args.filter,
                ),
                **self.serialize_query_param(
                    "api-version", "2026-01-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.next_link = AAZStrType(
                serialized_name="nextLink",
            )
            _schema_on_200.value = AAZListType(
                flags={"required": True},
            )

            value = cls._schema_on_200.value
            value.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element
            _element.id = AAZStrType(
                flags={"read_only": True},
            )
            _element.name = AAZStrType(
                flags={"read_only": True},
            )
            _element.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _element.system_data = AAZObjectType(
                serialized_name="systemData",
                flags={"read_only": True},
            )
            _element.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.value.Element.properties
            properties.assignment_scope_validation = AAZStrType(
                serialized_name="assignmentScopeValidation",
            )
            properties.description = AAZStrType()
            properties.display_name = AAZStrType(
                serialized_name="displayName",
            )
            properties.exemption_category = AAZStrType(
                serialized_name="exemptionCategory",
                flags={"required": True},
            )
            properties.expires_on = AAZStrType(
                serialized_name="expiresOn",
            )
            properties.metadata = AAZAnyType()
            properties.policy_assignment_id = AAZStrType(
                serialized_name="policyAssignmentId",
                flags={"required": True},
            )
            properties.policy_definition_reference_ids = AAZListType(
                serialized_name="policyDefinitionReferenceIds",
            )
            properties.resource_selectors = AAZListType(
                serialized_name="resourceSelectors",
            )

            policy_definition_reference_ids = cls._schema_on_200.value.Element.properties.policy_definition_reference_ids
            policy_definition_reference_ids.Element = AAZStrType()

            resource_selectors = cls._schema_on_200.value.Element.properties.resource_selectors
            resource_selectors.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element.properties.resource_selectors.Element
            _element.name = AAZStrType()
            _element.selectors = AAZListType()

            selectors = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors
            selectors.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element
            _element["in"] = AAZListType()
            _element.kind = AAZStrType()
            _element.not_in = AAZListType(
                serialized_name="notIn",
            )

            in_ = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element["in"]
            in_.Element = AAZStrType()

            not_in = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element.not_in
            not_in.Element = AAZStrType()

            system_data = cls._schema_on_200.value.Element.system_data
            system_data.created_at = AAZStrType(
                serialized_name="createdAt",
            )
            system_data.created_by = AAZStrType(
                serialized_name="createdBy",
            )
            system_data.created_by_type = AAZStrType(
                serialized_name="createdByType",
            )
            system_data.last_modified_at = AAZStrType(
                serialized_name="lastModifiedAt",
            )
            system_data.last_modified_by = AAZStrType(
                serialized_name="lastModifiedBy",
            )
            system_data.last_modified_by_type = AAZStrType(
                serialized_name="lastModifiedByType",
            )

            return cls._schema_on_200

    class PolicyExemptionsList(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/providers/Microsoft.Authorization/policyExemptions",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "$filter", self.ctx.args.filter,
                ),
                **self.serialize_query_param(
                    "api-version", "2026-01-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.next_link = AAZStrType(
                serialized_name="nextLink",
            )
            _schema_on_200.value = AAZListType(
                flags={"required": True},
            )

            value = cls._schema_on_200.value
            value.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element
            _element.id = AAZStrType(
                flags={"read_only": True},
            )
            _element.name = AAZStrType(
                flags={"read_only": True},
            )
            _element.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _element.system_data = AAZObjectType(
                serialized_name="systemData",
                flags={"read_only": True},
            )
            _element.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.value.Element.properties
            properties.assignment_scope_validation = AAZStrType(
                serialized_name="assignmentScopeValidation",
            )
            properties.description = AAZStrType()
            properties.display_name = AAZStrType(
                serialized_name="displayName",
            )
            properties.exemption_category = AAZStrType(
                serialized_name="exemptionCategory",
                flags={"required": True},
            )
            properties.expires_on = AAZStrType(
                serialized_name="expiresOn",
            )
            properties.metadata = AAZAnyType()
            properties.policy_assignment_id = AAZStrType(
                serialized_name="policyAssignmentId",
                flags={"required": True},
            )
            properties.policy_definition_reference_ids = AAZListType(
                serialized_name="policyDefinitionReferenceIds",
            )
            properties.resource_selectors = AAZListType(
                serialized_name="resourceSelectors",
            )

            policy_definition_reference_ids = cls._schema_on_200.value.Element.properties.policy_definition_reference_ids
            policy_definition_reference_ids.Element = AAZStrType()

            resource_selectors = cls._schema_on_200.value.Element.properties.resource_selectors
            resource_selectors.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element.properties.resource_selectors.Element
            _element.name = AAZStrType()
            _element.selectors = AAZListType()

            selectors = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors
            selectors.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element
            _element["in"] = AAZListType()
            _element.kind = AAZStrType()
            _element.not_in = AAZListType(
                serialized_name="notIn",
            )

            in_ = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element["in"]
            in_.Element = AAZStrType()

            not_in = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element.not_in
            not_in.Element = AAZStrType()

            system_data = cls._schema_on_200.value.Element.system_data
            system_data.created_at = AAZStrType(
                serialized_name="createdAt",
            )
            system_data.created_by = AAZStrType(
                serialized_name="createdBy",
            )
            system_data.created_by_type = AAZStrType(
                serialized_name="createdByType",
            )
            system_data.last_modified_at = AAZStrType(
                serialized_name="lastModifiedAt",
            )
            system_data.last_modified_by = AAZStrType(
                serialized_name="lastModifiedBy",
            )
            system_data.last_modified_by_type = AAZStrType(
                serialized_name="lastModifiedByType",
            )

            return cls._schema_on_200

    class PolicyExemptionsListForResourceGroup(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Authorization/policyExemptions",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "$filter", self.ctx.args.filter,
                ),
                **self.serialize_query_param(
                    "api-version", "2026-01-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.next_link = AAZStrType(
                serialized_name="nextLink",
            )
            _schema_on_200.value = AAZListType(
                flags={"required": True},
            )

            value = cls._schema_on_200.value
            value.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element
            _element.id = AAZStrType(
                flags={"read_only": True},
            )
            _element.name = AAZStrType(
                flags={"read_only": True},
            )
            _element.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _element.system_data = AAZObjectType(
                serialized_name="systemData",
                flags={"read_only": True},
            )
            _element.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.value.Element.properties
            properties.assignment_scope_validation = AAZStrType(
                serialized_name="assignmentScopeValidation",
            )
            properties.description = AAZStrType()
            properties.display_name = AAZStrType(
                serialized_name="displayName",
            )
            properties.exemption_category = AAZStrType(
                serialized_name="exemptionCategory",
                flags={"required": True},
            )
            properties.expires_on = AAZStrType(
                serialized_name="expiresOn",
            )
            properties.metadata = AAZAnyType()
            properties.policy_assignment_id = AAZStrType(
                serialized_name="policyAssignmentId",
                flags={"required": True},
            )
            properties.policy_definition_reference_ids = AAZListType(
                serialized_name="policyDefinitionReferenceIds",
            )
            properties.resource_selectors = AAZListType(
                serialized_name="resourceSelectors",
            )

            policy_definition_reference_ids = cls._schema_on_200.value.Element.properties.policy_definition_reference_ids
            policy_definition_reference_ids.Element = AAZStrType()

            resource_selectors = cls._schema_on_200.value.Element.properties.resource_selectors
            resource_selectors.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element.properties.resource_selectors.Element
            _element.name = AAZStrType()
            _element.selectors = AAZListType()

            selectors = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors
            selectors.Element = AAZObjectType()

            _element = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element
            _element["in"] = AAZListType()
            _element.kind = AAZStrType()
            _element.not_in = AAZListType(
                serialized_name="notIn",
            )

            in_ = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element["in"]
            in_.Element = AAZStrType()

            not_in = cls._schema_on_200.value.Element.properties.resource_selectors.Element.selectors.Element.not_in
            not_in.Element = AAZStrType()

            system_data = cls._schema_on_200.value.Element.system_data
            system_data.created_at = AAZStrType(
                serialized_name="createdAt",
            )
            system_data.created_by = AAZStrType(
                serialized_name="createdBy",
            )
            system_data.created_by_type = AAZStrType(
                serialized_name="createdByType",
            )
            system_data.last_modified_at = AAZStrType(
                serialized_name="lastModifiedAt",
            )
            system_data.last_modified_by = AAZStrType(
                serialized_name="lastModifiedBy",
            )
            system_data.last_modified_by_type = AAZStrType(
                serialized_name="lastModifiedByType",
            )

            return cls._schema_on_200


class _ListHelper:
    """Helper class for List"""


__all__ = ["List"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from types import SimpleNamespace
from unittest.mock import Mock
from azure.cli.core.azclierror import InvalidArgumentValueError
from azure.cli.core.cloud import HARD_CODED_CLOUD_LIST
from azure.cli.command_modules.acr._validators import validate_registry_name

class AcrValidatorsTests(unittest.TestCase):  
    def test_registry_name_with_dnl_suffix_loginserver(self):
        namespace = SimpleNamespace(
            **{
                "registry_name": "myacr-dnlhash123.azurecr.io",
            }
        )
        azure_public_cloud = HARD_CODED_CLOUD_LIST[0]
        cmd = Mock(cli_ctx=Mock(cloud=azure_public_cloud))
        validate_registry_name(cmd, namespace)
        self.assertEqual(namespace.registry_name, 'myacr')

    def test_registry_name_valid_inputs(self):
        """Test valid registry names that should pass validation."""
        test_cases = [
            # (input_registry_name, expected_output)
            ("validregistry.azurecr.io", "validregistry"),
            ("myregistry123-dnlhasd234.azurecr.io", "myregistry123"),
            ("test5chars", "test5chars"),
            ("a" * 50, "a" * 50),  # Maximum length
            ("registry2024", "registry2024"),
        ]
        
        azure_public_cloud = HARD_CODED_CLOUD_LIST[0]
        cmd = Mock(cli_ctx=Mock(cloud=azure_public_cloud))
        
        for input_name, expected_output in test_cases:
            with self.subTest(input_name=input_name):
                namespace = SimpleNamespace(registry_name=input_name)
                validate_registry_name(cmd, namespace)
                self.assertEqual(namespace.registry_name, expected_output)
    
    def test_registry_name_invalid_inputs_should_raise_error(self):
        """Test invalid registry names that should raise InvalidArgumentValueError."""
        invalid_inputs = [
            "myregistry-hash123",  # Has hyphen but no suffix
            "test.invalid.suffix",  # Invalid suffix
            "registry.azurecr.io124567",  # Wrong suffix
            "my-registry.wrongsuffix.io",  # Invalid suffix with hyphen
            "78787%^&*(()).azurecr.io" # invalid characters
        ]
        
        azure_public_cloud = HARD_CODED_CLOUD_LIST[0]
        cmd = Mock(cli_ctx=Mock(cloud=azure_public_cloud))
        
        for invalid_input in invalid_inputs:
            with self.subTest(invalid_input=invalid_input):
                namespace = SimpleNamespace(registry_name=invalid_input)
                with self.assertRaises(InvalidArgumentValueError):
                    validate_registry_name(cmd, namespace)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core import _debug
from azure.cli.core.auth.util import resource_to_scopes
from azure.cli.core.extension import EXTENSIONS_MOD_PREFIX
from azure.cli.core.profiles import ResourceType, CustomResourceType, get_api_version
from azure.cli.core.profiles._shared import get_client_class, SDKProfile
from azure.cli.core.util import get_az_user_agent, is_track2
from knack.log import get_logger
from knack.util import CLIError

logger = get_logger(__name__)


def _is_vendored_sdk_path(path_comps):
    return len(path_comps) >= 5 and path_comps[4] == 'vendored_sdks'


def resolve_client_arg_name(operation, kwargs):
    if not isinstance(operation, str):
        raise CLIError("operation should be type 'str'. Got '{}'".format(type(operation)))
    if 'client_arg_name' in kwargs:
        logger.info("Keyword 'client_arg_name' is deprecated and should be removed.")
        return kwargs['client_arg_name']
    path, op_path = operation.split('#', 1)

    path_comps = path.split('.')
    if path_comps[0] == 'azure':
        if path_comps[1] != 'cli' or _is_vendored_sdk_path(path_comps):
            # Public SDK: azure.mgmt.resource... (mgmt-plane) or azure.storage.blob... (data-plane)
            # Vendored SDK: azure.cli.command_modules.keyvault.vendored_sdks...
            client_arg_name = 'self'
        else:
            # CLI custom method: azure.cli.command_modules.resource...
            client_arg_name = 'client'
    elif path_comps[0].startswith(EXTENSIONS_MOD_PREFIX):
        # for CLI extensions
        # SDK method: the operation takes the form '<class name>.<method_name>'
        # custom method: the operation takes the form '<method_name>'
        op_comps = op_path.split('.')
        client_arg_name = 'self' if len(op_comps) > 1 else 'client'
    else:
        raise ValueError('Unrecognized operation: {}'.format(operation))
    return client_arg_name


def get_mgmt_service_client(cli_ctx, client_or_resource_type, subscription_id=None, api_version=None,
                            aux_subscriptions=None, aux_tenants=None, credential=None, **kwargs):
    """Create Python SDK mgmt-plane client. In addition to directly creating the client instance, this centralized
    client factory adds more features to the client instance:
    - multi-API
    - server telemetry
    - safe logging
    - cross-tenant authentication

    :param cli_ctx: AzCli instance
    :param client_or_resource_type: Python SDK mgmt-plane client type or a member of
        azure.cli.core.profiles._shared.ResourceType
    :param subscription_id: Override the current login context's subscription
    :param api_version: Override the client or resource type's default API version
    :param aux_subscriptions: For cross-tenant authentication, such as vnet peering
    :param aux_tenants: For cross-tenant authentication, such as vnet peering
    :param credential: Custom credential to override the current login context's credential
        For more information, see src/azure-cli-core/azure/cli/core/auth/msal_authentication.py
    """
    if not subscription_id and 'subscription_id' in cli_ctx.data:
        subscription_id = cli_ctx.data['subscription_id']

    sdk_profile = None
    if isinstance(client_or_resource_type, (ResourceType, CustomResourceType)):
        # Get the versioned client
        client_type = get_client_class(client_or_resource_type)
        api_version = api_version or get_api_version(cli_ctx, client_or_resource_type, as_sdk_profile=True)
        if isinstance(api_version, SDKProfile):
            sdk_profile = api_version.profile
            api_version = None
    else:
        # Get the non-versioned client
        client_type = client_or_resource_type
    client, _ = _get_mgmt_service_client(cli_ctx, client_type, subscription_id=subscription_id,
                                         api_version=api_version, sdk_profile=sdk_profile,
                                         aux_subscriptions=aux_subscriptions,
                                         aux_tenants=aux_tenants, credential=credential,
                                         **kwargs)
    return client


def get_subscription_service_client(cli_ctx):
    return _get_mgmt_service_client(cli_ctx, get_client_class(ResourceType.MGMT_RESOURCE_SUBSCRIPTIONS),
                                    subscription_bound=False,
                                    api_version=get_api_version(cli_ctx, ResourceType.MGMT_RESOURCE_SUBSCRIPTIONS))


def configure_common_settings(cli_ctx, client):
    client = _debug.change_ssl_cert_verification(client)

    client.config.enable_http_logger = True

    client.config.add_user_agent(get_az_user_agent())

    try:
        command_ext_name = cli_ctx.data['command_extension_name']
        if command_ext_name:
            client.config.add_user_agent("CliExtension/{}".format(command_ext_name))
    except KeyError:
        pass

    # Prepare CommandName header
    command_name_suffix = ';completer-request' if cli_ctx.data['completer_active'] else ''
    cli_ctx.data['headers']['CommandName'] = "{}{}".format(cli_ctx.data['command'], command_name_suffix)

    # Prepare ParameterSetName header
    if cli_ctx.data.get('safe_params'):
        cli_ctx.data['headers']['ParameterSetName'] = ' '.join(cli_ctx.data['safe_params'])

    # Prepare x-ms-client-request-id header
    client.config.generate_client_request_id = 'x-ms-client-request-id' not in cli_ctx.data['headers']

    logger.debug("Adding custom headers to the client:")

    for header, value in cli_ctx.data['headers'].items():
        # msrest doesn't print custom headers in debug log, so CLI should do that
        logger.debug("    '%s': '%s'", header, value)
        # We are working with the autorest team to expose the add_header functionality of the generated client to avoid
        # having to access private members
        client._client.add_header(header, value)  # pylint: disable=protected-access


def _prepare_client_kwargs_track2(cli_ctx):
    """Prepare kwargs for Track 2 SDK client."""
    client_kwargs = {}

    # Prepare connection_verify to change SSL verification behavior, used by ConnectionConfiguration
    client_kwargs.update(_debug.change_ssl_cert_verification_track2())

    # Prepare User-Agent header, used by UserAgentPolicy
    client_kwargs['user_agent'] = get_az_user_agent()

    try:
        command_ext_name = cli_ctx.data['command_extension_name']
        if command_ext_name:
            client_kwargs['user_agent'] += "CliExtension/{}".format(command_ext_name)
    except KeyError:
        pass

    # Prepare custom headers, used by HeadersPolicy
    headers = dict(cli_ctx.data['headers'])

    # - Prepare CommandName header
    command_name_suffix = ';completer-request' if cli_ctx.data['completer_active'] else ''
    headers['CommandName'] = "{}{}".format(cli_ctx.data['command'], command_name_suffix)

    # - Prepare ParameterSetName header
    if cli_ctx.data.get('safe_params'):
        headers['ParameterSetName'] = ' '.join(cli_ctx.data['safe_params'])

    client_kwargs['headers'] = headers

    # Prepare x-ms-client-request-id header, used by RequestIdPolicy
    if 'x-ms-client-request-id' in cli_ctx.data['headers']:
        client_kwargs['request_id'] = cli_ctx.data['headers']['x-ms-client-request-id']

    from azure.cli.core.sdk.policies import RecordTelemetryUserAgentPolicy
    client_kwargs['user_agent_policy'] = RecordTelemetryUserAgentPolicy(**client_kwargs)

    from azure.cli.core.sdk.policies import get_custom_hook_policy
    client_kwargs['custom_hook_policy'] = get_custom_hook_policy(cli_ctx)

    # Replace NetworkTraceLoggingPolicy to redact 'Authorization' and 'x-ms-authorization-auxiliary' headers.
    #   NetworkTraceLoggingPolicy: log raw network trace, with all headers.
    from azure.cli.core.sdk.policies import SafeNetworkTraceLoggingPolicy
    client_kwargs['logging_policy'] = SafeNetworkTraceLoggingPolicy()

    # Disable ARMHttpLoggingPolicy.
    #   ARMHttpLoggingPolicy: Only log allowed information.
    from azure.core.pipeline.policies import SansIOHTTPPolicy
    client_kwargs['http_logging_policy'] = SansIOHTTPPolicy()

    return client_kwargs


def _prepare_mgmt_client_kwargs_track2(cli_ctx, cred):
    """Prepare kwargs for Track 2 SDK mgmt client."""
    client_kwargs = _prepare_client_kwargs_track2(cli_ctx)

    # Enable CAE support in mgmt SDK
    from azure.mgmt.core.policies import ARMChallengeAuthenticationPolicy

    # Track 2 SDK maintains `scopes` and passes `scopes` to get_token.
    scopes = resource_to_scopes(cli_ctx.cloud.endpoints.active_directory_resource_id)
    policy = ARMChallengeAuthenticationPolicy(cred, *scopes)

    client_kwargs['credential_scopes'] = scopes
    client_kwargs['authentication_policy'] = policy

    # Track 2 currently lacks the ability to take external credentials.
    #   https://github.com/Azure/azure-sdk-for-python/issues/8313
    # As a temporary workaround, manually add external tokens to 'x-ms-authorization-auxiliary' header.
    #   https://learn.microsoft.com/en-us/azure/azure-resource-manager/management/authenticate-multi-tenant
    if hasattr(cred, "get_auxiliary_tokens"):
        aux_tokens = cred.get_auxiliary_tokens(*scopes)
        if aux_tokens:
            # Hard-code scheme to 'Bearer' as _BearerTokenCredentialPolicyBase._update_headers does.
            client_kwargs['headers']['x-ms-authorization-auxiliary'] = \
                ', '.join("Bearer {}".format(token.token) for token in aux_tokens)

    return client_kwargs


def _get_mgmt_service_client(cli_ctx,
                             client_type,
                             subscription_bound=True,
                             subscription_id=None,
                             api_version=None,
                             base_url_bound=True,
                             sdk_profile=None,
                             aux_subscriptions=None,
                             aux_tenants=None,
                             credential=None,
                             **kwargs):
    from azure.cli.core._profile import Profile
    logger.debug('Getting management service client client_type=%s', client_type.__name__)

    if credential:
        # Use a custom credential
        if not subscription_id:
            raise ValueError('credential and subscription_id must be specified at the same time.')
    else:
        # Get a credential for the current `az login` context
        profile = Profile(cli_ctx=cli_ctx)
        credential, subscription_id, _ = profile.get_login_credentials(
            subscription_id=subscription_id,
            aux_subscriptions=aux_subscriptions, aux_tenants=aux_tenants)

    client_kwargs = {}
    if base_url_bound:
        client_kwargs = {'base_url': cli_ctx.cloud.endpoints.resource_manager}
    if api_version:
        client_kwargs['api_version'] = api_version
    if sdk_profile:
        client_kwargs['profile'] = sdk_profile
    if kwargs:
        client_kwargs.update(kwargs)

    if is_track2(client_type):
        client_kwargs.update(_prepare_mgmt_client_kwargs_track2(cli_ctx, credential))

    if subscription_bound:
        client = client_type(credential, subscription_id, **client_kwargs)
    else:
        client = client_type(credential, **client_kwargs)

    if not is_track2(client):
        configure_common_settings(cli_ctx, client)

    return client, subscription_id


def get_data_service_client(cli_ctx, service_type, account_name, account_key, connection_string=None,
                            sas_token=None, socket_timeout=None, token_credential=None, endpoint_suffix=None,
                            location_mode=None):
    logger.debug('Getting data service client service_type=%s', service_type.__name__)
    try:
        client_kwargs = {'account_name': account_name,
                         'account_key': account_key,
                         'connection_string': connection_string,
                         'sas_token': sas_token}
        if socket_timeout:
            client_kwargs['socket_timeout'] = socket_timeout
        if token_credential:
            client_kwargs['token_credential'] = token_credential
        if endpoint_suffix:
            client_kwargs['endpoint_suffix'] = endpoint_suffix
        client = service_type(**client_kwargs)
        if location_mode:
            client.location_mode = location_mode
    except ValueError as exc:
        _ERROR_STORAGE_MISSING_INFO = ('You need to provide an account name and either an account_key or sas_token '
                                       'when creating a storage service.')
        if _ERROR_STORAGE_MISSING_INFO in str(exc):
            raise ValueError(exc)
        raise CLIError('Unable to obtain data client. Check your connection parameters.')
    # TODO: enable Fiddler
    client.request_callback = _get_add_headers_callback(cli_ctx)
    return client


def get_subscription_id(cli_ctx):
    from azure.cli.core._profile import Profile
    if not cli_ctx.data.get('subscription_id'):
        cli_ctx.data['subscription_id'] = Profile(cli_ctx=cli_ctx).get_subscription_id()
    return cli_ctx.data['subscription_id']


def _get_add_headers_callback(cli_ctx):

    def _add_headers(request):
        agents = [request.headers['User-Agent'], get_az_user_agent()]
        request.headers['User-Agent'] = ' '.join(agents)

        try:
            request.headers.update(cli_ctx.data['headers'])
        except KeyError:
            pass

    return _add_headers


prepare_client_kwargs_track2 = _prepare_client_kwargs_track2



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "policy assignment identity show",
)
class Show(AAZCommand):
    """Retrieve the managed identity.

    Retrieve and show the details of the system or user assigned managed identity of the policy assignment matching the given name and scope.

    :example: Show a policy assignment's managed identity
        az policy assignment identity show --name MyPolicyAssignment --scope '/providers/Microsoft.Management/managementGroups/{managementGroupName}'
    """

    _aaz_info = {
        "version": "2025-11-01",
        "resources": [
            ["mgmt-plane", "/{scope}/providers/microsoft.authorization/policyassignments/{}", "2025-11-01", "identity"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self.SubresourceSelector(ctx=self.ctx, name="subresource")
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.name = AAZStrArg(
            options=["-n", "--name"],
            help={"short-summary": "The name of the policy assignment.", "long-summary": "The name of the policy assignment is the name segment of its resource ID."},
            required=True,
            fmt=AAZStrArgFormat(
                pattern="^[^<>%&:\\?/]*[^<>%&:\\?/ ]+$",
            ),
        )
        _args_schema.scope = AAZStrArg(
            options=["--scope"],
            help={"short-summary": "The scope of the policy assignment.", "long-summary": "Valid scopes are: management group (format: '/providers/Microsoft.Management/managementGroups/{managementGroup}'), subscription (format: '/subscriptions/{subscriptionId}'), resource group (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}', or resource (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/[{parentResourcePath}/]{resourceType}/{resourceName}'. The scope of an assignment is always the part of its ID preceding '/providers/Microsoft.Authorization/policyAssignments/{policyAssignmentName}'. If scope is not provided, the scope will be the implied or specified subscription."},
            required=True,
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        self.PolicyAssignmentsGet(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.selectors.subresource.required(), client_flatten=True)
        return result

    class SubresourceSelector(AAZJsonSelector):

        def _get(self):
            result = self.ctx.vars.instance
            return result.identity

        def _set(self, value):
            result = self.ctx.vars.instance
            result.identity = value
            return

    class PolicyAssignmentsGet(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/{scope}/providers/Microsoft.Authorization/policyAssignments/{policyAssignmentName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "policyAssignmentName", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "scope", self.ctx.args.scope,
                    skip_quote=True,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2025-11-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()
            _ShowHelper._build_schema_policy_assignment_read(cls._schema_on_200)

            return cls._schema_on_200


class _ShowHelper:
    """Helper class for Show"""

    _schema_policy_assignment_read = None

    @classmethod
    def _build_schema_policy_assignment_read(cls, _schema):
        if cls._schema_policy_assignment_read is not None:
            _schema.id = cls._schema_policy_assignment_read.id
            _schema.identity = cls._schema_policy_assignment_read.identity
            _schema.location = cls._schema_policy_assignment_read.location
            _schema.name = cls._schema_policy_assignment_read.name
            _schema.properties = cls._schema_policy_assignment_read.properties
            _schema.system_data = cls._schema_policy_assignment_read.system_data
            _schema.type = cls._schema_policy_assignment_read.type
            return

        cls._schema_policy_assignment_read = _schema_policy_assignment_read = AAZObjectType()

        policy_assignment_read = _schema_policy_assignment_read
        policy_assignment_read.id = AAZStrType(
            flags={"read_only": True},
        )
        policy_assignment_read.identity = AAZIdentityObjectType()
        policy_assignment_read.location = AAZStrType()
        policy_assignment_read.name = AAZStrType(
            flags={"read_only": True},
        )
        policy_assignment_read.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        policy_assignment_read.system_data = AAZObjectType(
            serialized_name="systemData",
            flags={"read_only": True},
        )
        policy_assignment_read.type = AAZStrType(
            flags={"read_only": True},
        )

        identity = _schema_policy_assignment_read.identity
        identity.principal_id = AAZStrType(
            serialized_name="principalId",
            flags={"read_only": True},
        )
        identity.tenant_id = AAZStrType(
            serialized_name="tenantId",
            flags={"read_only": True},
        )
        identity.type = AAZStrType()
        identity.user_assigned_identities = AAZDictType(
            serialized_name="userAssignedIdentities",
        )

        user_assigned_identities = _schema_policy_assignment_read.identity.user_assigned_identities
        user_assigned_identities.Element = AAZObjectType()

        _element = _schema_policy_assignment_read.identity.user_assigned_identities.Element
        _element.client_id = AAZStrType(
            serialized_name="clientId",
            flags={"read_only": True},
        )
        _element.principal_id = AAZStrType(
            serialized_name="principalId",
            flags={"read_only": True},
        )

        properties = _schema_policy_assignment_read.properties
        properties.assignment_type = AAZStrType(
            serialized_name="assignmentType",
        )
        properties.definition_version = AAZStrType(
            serialized_name="definitionVersion",
        )
        properties.description = AAZStrType()
        properties.display_name = AAZStrType(
            serialized_name="displayName",
        )
        properties.effective_definition_version = AAZStrType(
            serialized_name="effectiveDefinitionVersion",
            flags={"read_only": True},
        )
        properties.enforcement_mode = AAZStrType(
            serialized_name="enforcementMode",
        )
        properties.instance_id = AAZStrType(
            serialized_name="instanceId",
            flags={"read_only": True},
        )
        properties.latest_definition_version = AAZStrType(
            serialized_name="latestDefinitionVersion",
            flags={"read_only": True},
        )
        properties.metadata = AAZAnyType()
        properties.non_compliance_messages = AAZListType(
            serialized_name="nonComplianceMessages",
        )
        properties.not_scopes = AAZListType(
            serialized_name="notScopes",
        )
        properties.overrides = AAZListType()
        properties.parameters = AAZDictType()
        properties.policy_definition_id = AAZStrType(
            serialized_name="policyDefinitionId",
        )
        properties.resource_selectors = AAZListType(
            serialized_name="resourceSelectors",
        )
        properties.scope = AAZStrType(
            flags={"read_only": True},
        )
        properties.self_serve_exemption_settings = AAZObjectType(
            serialized_name="selfServeExemptionSettings",
        )

        non_compliance_messages = _schema_policy_assignment_read.properties.non_compliance_messages
        non_compliance_messages.Element = AAZObjectType()

        _element = _schema_policy_assignment_read.properties.non_compliance_messages.Element
        _element.message = AAZStrType(
            flags={"required": True},
        )
        _element.policy_definition_reference_id = AAZStrType(
            serialized_name="policyDefinitionReferenceId",
        )

        not_scopes = _schema_policy_assignment_read.properties.not_scopes
        not_scopes.Element = AAZStrType()

        overrides = _schema_policy_assignment_read.properties.overrides
        overrides.Element = AAZObjectType()

        _element = _schema_policy_assignment_read.properties.overrides.Element
        _element.kind = AAZStrType()
        _element.selectors = AAZListType()
        _element.value = AAZStrType()

        selectors = _schema_policy_assignment_read.properties.overrides.Element.selectors
        selectors.Element = AAZObjectType()
        cls._build_schema_selector_read(selectors.Element)

        parameters = _schema_policy_assignment_read.properties.parameters
        parameters.Element = AAZObjectType()

        _element = _schema_policy_assignment_read.properties.parameters.Element
        _element.value = AAZAnyType()

        resource_selectors = _schema_policy_assignment_read.properties.resource_selectors
        resource_selectors.Element = AAZObjectType()

        _element = _schema_policy_assignment_read.properties.resource_selectors.Element
        _element.name = AAZStrType()
        _element.selectors = AAZListType()

        selectors = _schema_policy_assignment_read.properties.resource_selectors.Element.selectors
        selectors.Element = AAZObjectType()
        cls._build_schema_selector_read(selectors.Element)

        self_serve_exemption_settings = _schema_policy_assignment_read.properties.self_serve_exemption_settings
        self_serve_exemption_settings.enabled = AAZBoolType()
        self_serve_exemption_settings.policy_definition_reference_ids = AAZListType(
            serialized_name="policyDefinitionReferenceIds",
        )

        policy_definition_reference_ids = _schema_policy_assignment_read.properties.self_serve_exemption_settings.policy_definition_reference_ids
        policy_definition_reference_ids.Element = AAZStrType()

        system_data = _schema_policy_assignment_read.system_data
        system_data.created_at = AAZStrType(
            serialized_name="createdAt",
        )
        system_data.created_by = AAZStrType(
            serialized_name="createdBy",
        )
        system_data.created_by_type = AAZStrType(
            serialized_name="createdByType",
        )
        system_data.last_modified_at = AAZStrType(
            serialized_name="lastModifiedAt",
        )
        system_data.last_modified_by = AAZStrType(
            serialized_name="lastModifiedBy",
        )
        system_data.last_modified_by_type = AAZStrType(
            serialized_name="lastModifiedByType",
        )

        _schema.id = cls._schema_policy_assignment_read.id
        _schema.identity = cls._schema_policy_assignment_read.identity
        _schema.location = cls._schema_policy_assignment_read.location
        _schema.name = cls._schema_policy_assignment_read.name
        _schema.properties = cls._schema_policy_assignment_read.properties
        _schema.system_data = cls._schema_policy_assignment_read.system_data
        _schema.type = cls._schema_policy_assignment_read.type

    _schema_selector_read = None

    @classmethod
    def _build_schema_selector_read(cls, _schema):
        if cls._schema_selector_read is not None:
            _schema["in"] = cls._schema_selector_read["in"]
            _schema.kind = cls._schema_selector_read.kind
            _schema.not_in = cls._schema_selector_read.not_in
            return

        cls._schema_selector_read = _schema_selector_read = AAZObjectType()

        selector_read = _schema_selector_read
        selector_read["in"] = AAZListType()
        selector_read.kind = AAZStrType()
        selector_read.not_in = AAZListType(
            serialized_name="notIn",
        )

        in_ = _schema_selector_read["in"]
        in_.Element = AAZStrType()

        not_in = _schema_selector_read.not_in
        not_in.Element = AAZStrType()

        _schema["in"] = cls._schema_selector_read["in"]
        _schema.kind = cls._schema_selector_read.kind
        _schema.not_in = cls._schema_selector_read.not_in


__all__ = ["Show"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import time
from datetime import datetime, timezone
from azure.cli.testsdk.scenario_tests.const import ENV_LIVE_TEST
from azure.cli.testsdk.scenario_tests import AllowLargeResponse
from azure.cli.testsdk import (
    JMESPathCheck,
    NoneCheck,
    ResourceGroupPreparer,
    ScenarioTest)
from .constants import DEFAULT_LOCATION, SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH


def _unique_tag_subset_check(tags):
    unique_tags = {}
    for tag in tags.split():
        key, value = tag.split('=', 1)
        unique_tags[key] = value

    if not unique_tags:
        return JMESPathCheck('tags', {})

    subset_query = 'tags.{{{}}}'.format(', '.join('{0}: {0}'.format(key) for key in unique_tags))
    return JMESPathCheck(subset_query, unique_tags)


class PostgreSQLFlexibleServerTagsMgmtScenarioTest(ScenarioTest):

    postgres_location = DEFAULT_LOCATION

    @AllowLargeResponse()
    @ResourceGroupPreparer(location=postgres_location)
    def test_postgres_flexible_server_tags_mgmt(self, resource_group):
        self._test_flexible_server_tags_mgmt(resource_group)

    def _test_flexible_server_tags_mgmt(self, resource_group):

        initial_tags = "tag1=value1 tag2=value2 tag3=value3"
        new_tags = "tag4=value4 tag5=value5"
        new_repeated_tags = "tag6=value6a tag6=value6b tag7=value7"
        # Following two are defined for the future, for when backend supports properly using passed tags to restore and revive operations.
        restored_server_tags = "tag8=value8 tag9=value9"
        revived_server_tags = "tag10=value10 tag11=value11"

        # Create server.
        if self.cli_ctx.local_context.is_on:
            self.cmd('config param-persist off')

        location = self.postgres_location

        primary_server = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)
        first_level_replica_1 = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)
        first_level_replica_2 = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)
        cascade_replica_1 = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)
        cascade_replica_2 = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)
        restored_server_1 = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)
        restored_server_2 = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)
        revive_dropped_server = self.create_random_name(SERVER_NAME_PREFIX, SERVER_NAME_MAX_LENGTH)

        # Create server with three tags.
        self.cmd('postgres flexible-server create -g {} -n {} -l {} --tags {} --public-access enabled --yes'
                 .format(resource_group, primary_server, location, initial_tags))

        # Validate that tags are added to the created server.
        self.cmd('postgres flexible-server show -g {} -n {}'
                                 .format(resource_group, primary_server),
                                 checks=[_unique_tag_subset_check(initial_tags)])

        # Update server tags with two new tags, and validate that the original tags are removed and only the new tags are present on the server.
        self.cmd('postgres flexible-server update -g {} -n {} --tags {}'
                 .format(resource_group, primary_server, new_tags),
                 checks=[_unique_tag_subset_check(new_tags)])

        # Update server tags with two new repeated tags, and validate that only one of the repeated tags is present along with the other new tag, while all original tags are removed from the server.
        self.cmd('postgres flexible-server update -g {} -n {} --tags {}'
                 .format(resource_group, primary_server, new_repeated_tags),
                 checks=[_unique_tag_subset_check(new_repeated_tags)])

        # Stop server.
        self.cmd('postgres flexible-server stop -g {} -n {}'.format(resource_group, primary_server))

        # Update server tags with original tags while server is stopped. Not supported for flexible server. Expect failure.
        self.cmd('postgres flexible-server update -g {} -n {} --tags {}'
                 .format(resource_group, primary_server, initial_tags), expect_failure=True)

        # Start server.
        self.cmd('postgres flexible-server start -g {} -n {}'.format(resource_group, primary_server))

        # Update server tags with original tags. Should succeed now that server is started.
        self.cmd('postgres flexible-server update -g {} -n {} --tags {}'
                 .format(resource_group, primary_server, initial_tags),
                 checks=[_unique_tag_subset_check(initial_tags)])
        
        # Add a first level read replica to the server and don't provide tags, to validate that tags are inherited from its source server.
        self.cmd('postgres flexible-server replica create -g {} -n {} --source-server {} --location {} --yes'
                 .format(resource_group, first_level_replica_1, primary_server, location),
                 checks=[_unique_tag_subset_check(initial_tags)])

        # Add a cascade read replica to the first read replica and provide tags, to validate that these provided tags are properly set on cascade replica.
        self.cmd('postgres flexible-server replica create -g {} -n {} --source-server {} --location {} --tags {} --yes'
                 .format(resource_group, cascade_replica_1, first_level_replica_1, location, new_repeated_tags),
                 checks=[_unique_tag_subset_check(new_repeated_tags)])

        # Add a first level read replica to the server and provide tags, to validate that these provided tags are properly set on the first level replica.
        self.cmd('postgres flexible-server replica create -g {} -n {} --source-server {} --location {} --tags {} --yes'
                 .format(resource_group, first_level_replica_2, primary_server, location, new_repeated_tags),
                 checks=[_unique_tag_subset_check(new_repeated_tags)])

        # Add a cascade read replica to the second read replica and provide tags, to validate that these provided tags are properly set on the cascade replica.
        self.cmd('postgres flexible-server replica create -g {} -n {} --source-server {} --location {} --tags {} --yes'
                 .format(resource_group, cascade_replica_2, first_level_replica_2, location, ""),
                 checks=[_unique_tag_subset_check("")])
        
        # Confirm that primary server has at least one backup before testing restore with tags, as restore operation requires a backup to be present.
        self.cmd('postgres flexible-server backup list -g {} -s {}'.format(resource_group, primary_server),
                 checks=[JMESPathCheck("length(@) >= `1`", True)])

        restore_time_utc = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%SZ')

        # Restore server from backup without specific tags, to validate that tags on the restored server are inherited from those on the source server.
        self.cmd('postgres flexible-server restore -g {} -n {} --source-server {} --restore-time {} --yes'
                 .format(resource_group, restored_server_1, primary_server, restore_time_utc),
                 checks=[_unique_tag_subset_check(initial_tags)])
        
        # Update server tags with empty tags.
        self.cmd('postgres flexible-server update -g {} -n {} --tags {}'
                 .format(resource_group, primary_server, ""),
                 checks=[_unique_tag_subset_check("")])

        # Restore server from backup without specific tags, to validate that tags on the restored server are inherited from those on the source server.
        self.cmd('postgres flexible-server restore -g {} -n {} --source-server {} --restore-time {} --yes'
                 .format(resource_group, restored_server_2, primary_server, restore_time_utc),
                 checks=[_unique_tag_subset_check("")])

        # Update server tags with original tags.
        self.cmd('postgres flexible-server update -g {} -n {} --tags {}'
                 .format(resource_group, primary_server, initial_tags),
                 checks=[_unique_tag_subset_check(initial_tags)])
        
        # Fetch primary server resource identifier to be used as source server for revive operation.
        primary_server_resource_id = self.cmd('postgres flexible-server show -g {} -n {} --query id -o tsv'
                                            .format(resource_group, primary_server)).output.strip()

        # Delete servers in dependency order: deepest replicas first, then their source servers.
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, cascade_replica_1),
                 checks=NoneCheck())
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, cascade_replica_2),
                 checks=NoneCheck())
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, first_level_replica_1),
                 checks=NoneCheck())
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, first_level_replica_2),
                 checks=NoneCheck())
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, primary_server),
                 checks=NoneCheck())
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, restored_server_1),
                 checks=NoneCheck())
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, restored_server_2),
                 checks=NoneCheck())

        # Sleep for 5 minutes to ensure that the deleted primary server is fully deleted before attempting revive operation, as revive operation on a server that is not fully deleted is not allowed.
        os.environ.get(ENV_LIVE_TEST, False) and time.sleep(5 * 60)

        # Revive dropped server without specific tags, to validate that tags on the revived server are inherited from those on the tombstoned server.
        self.cmd('postgres flexible-server revive-dropped -g {} -n {} --source-server {} --location {}'
                 .format(resource_group, revive_dropped_server, primary_server_resource_id, location),
                 checks=[_unique_tag_subset_check(initial_tags)])
        
        # Delete revived server to clean up after test.
        self.cmd('postgres flexible-server delete -g {} -n {} --yes'.format(resource_group, revive_dropped_server),
                 checks=NoneCheck())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=unused-argument, line-too-long
from azure.cli.core.azclierror import RequiredArgumentMissingError
from azure.cli.core.util import CLIError, user_confirmation
from knack.log import get_logger
from ..utils.validators import (
    validate_resource_group,
    check_resource_group,
    validate_citus_cluster,
    validate_database_name)

logger = get_logger(__name__)


def database_create_func(cmd, client, resource_group_name, server_name, database_name=None, charset=None, collation=None):
    validate_database_name(database_name)
    validate_resource_group(resource_group_name)
    validate_citus_cluster(cmd, resource_group_name, server_name)

    if charset is None and collation is None:
        charset = 'utf8'
        collation = 'en_US.utf8'
        logger.warning("Creating database with utf8 charset and en_US.utf8 collation")
    elif (not charset and collation) or (charset and not collation):
        raise RequiredArgumentMissingError("charset and collation have to be input together.")

    parameters = {
        'name': database_name,
        'charset': charset,
        'collation': collation
    }

    return client.begin_create(
        resource_group_name,
        server_name,
        database_name,
        parameters)


def database_delete_func(cmd, client, resource_group_name=None, server_name=None, database_name=None, yes=None):
    if not check_resource_group(resource_group_name):
        resource_group_name = None

    result = None
    if resource_group_name is None or server_name is None or database_name is None:
        raise CLIError("Incorrect Usage : Deleting a database needs --resource-group, --server-name and --name. "
                       "If your parameter persistence is turned ON, make sure these three parameters exist in "
                       "persistent parameters using \'az config param-persist show\'. "
                       "If your parameter persistence is turned OFF, consider passing them explicitly.")
    if not yes:
        user_confirmation(
            "Are you sure you want to delete the database '{0}' of server '{1}'".format(database_name,
                                                                                        server_name), yes=yes)

    validate_citus_cluster(cmd, resource_group_name, server_name)
    try:
        result = client.begin_delete(resource_group_name, server_name, database_name)
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(ex)
    return result
